# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
import jax.scipy as jsp

from netket.jax import tree_ravel
from netket.utils.api_utils import partial_from_kwargs
from netket.utils.deprecation import warn_deprecation


@partial_from_kwargs
def pinv_smooth(
    A,
    b,
    *,
    rtol: float = 1e-14,
    rtol_smooth: float = 1e-14,
    x0=None,
    rcond: float = None,
    rcond_smooth: float = None,
):
    r"""
    Solve the linear system by building a pseudo-inverse from the
    eigendecomposition obtained from :func:`jax.numpy.linalg.eigh`.

    The eigenvalues :math:`\lambda_i` smaller than
    :math:`r_\textrm{cond} \lambda_\textrm{max}` are truncated (where
    :math:`\lambda_\textrm{max}` is the largest eigenvalue).

    The eigenvalues are further smoothed with another filter, originally introduced in
    `Medvidovic, Sels arXiv:2212.11289 (2022) <https://arxiv.org/abs/2212.11289>`_,
    given by the following equation

    .. math::

        \tilde\lambda_i^{-1}=\frac{\lambda_i^{-1}}{1+\big(\epsilon\frac{\lambda_\textrm{max}}{\lambda_i}\big)^6}


    .. note::

        In general, we found that this custom implementation of
        the pseudo-inverse outperform
        jax's :func:`~jax.numpy.linalg.pinv`. This might be
        because :func:`~jax.numpy.linalg.pinv` internally calls
        :obj:`~jax.numpy.linalg.svd`, while this solver internally
        uses :obj:`~jax.numpy.linalg.eigh`.

        For that reason, we suggest you use this solver instead of
        :obj:`~netket.optimizer.solver.pinv`.


    .. note::

        If you pass only keyword arguments, this solver will directly create
        a partial capturing them.


    Args:
        A: LinearOperator (matrix)
        b: vector or Pytree
        rtol : Relative tolerance for small singular values of :code:`A`. For
            the purposes of rank determination, singular values are treated
            as zero if they are smaller than `rtol` times the largest
            singular value of :code:`A`.
        rtol_smooth : Regularization parameter used with a similar effect to `rtol`
            but with a softer curve. See :math:`\epsilon` in the formula
            above.
        rcond: (deprecated) Alias for `rtol`. Will be removed in a future release.
        rcond_smooth: (deprecated) Alias for `rtol_smooth`. Will be removed in a future release.
    """
    del x0

    if rcond is not None:
        warn_deprecation(
            "The 'rcond' argument is deprecated and will be removed in a future release. "
            "Please use 'rtol' instead."
        )
        rtol = rcond

    if rcond_smooth is not None:
        warn_deprecation(
            "The 'rcond_smooth' argument is deprecated and will be removed in a future release. "
            "Please use 'rtol_smooth' instead."
        )
        rtol_smooth = rcond

    if not isinstance(A, jax.Array):
        A = A.to_dense()
    b, unravel = tree_ravel(b)

    Σ, U = jnp.linalg.eigh(A)

    # Discard eigenvalues below numerical precision
    Σ_inv = jnp.where(jnp.abs(Σ / Σ[-1]) > rtol, jnp.reciprocal(Σ), 0.0)

    # Set regularizer for singular value cutoff
    regularizer = 1.0 / (1.0 + (rtol_smooth / jnp.abs(Σ / Σ[-1])) ** 6)

    Σ_inv = Σ_inv * regularizer

    x = U @ (Σ_inv * (U.conj().T @ b))

    return unravel(x), None


@partial_from_kwargs
def pinv(A, b, *, rtol: float = 1e-12, x0=None, rcond: float = None):
    """
    Solve the linear system using jax's implementation of the
    pseudo-inverse.

    Internally it calls :func:`~jax.numpy.linalg.pinv` which
    uses a :func:`~jax.numpy.linalg.svd` decomposition with
    the same value of **rtol**.

    .. note::

        In general, we found that our custom implementation of
        the pseudo-inverse
        :func:`netket.optimizer.solver.pinv_smooth` (which
        internally uses hermitian diagonaliation) outperform
        jax's :func:`~jax.numpy.linalg.pinv`.

        For that reason, we suggest to use
        :func:`~netket.optimizer.solver.pinv_smooth` instead of
        :obj:`~netket.optimizer.solver.pinv`.


    .. note::

        If you pass only keyword arguments, this solver will directly create
        a partial capturing them.

    The diagonal shift on the matrix can be 0 and the
    **rtol** variable can be used to truncate small
    eigenvalues.

    Args:
        A: the matrix A in Ax=b
        b: the vector b in Ax=b
        rtol: Cut-off ratio for small singular values of :code:`A`.
        rcond: (deprecated) Alias for `rtol`. Will be removed in a future release.
    """
    del x0

    if rcond is not None:
        warn_deprecation(
            "The 'rcond' argument is deprecated and will be removed in a future release. "
            "Please use 'rtol' instead."
        )
        rtol = rcond

    if not isinstance(A, jax.Array):
        A = A.to_dense()
    b, unravel = tree_ravel(b)

    A_inv = jnp.linalg.pinv(A, rtol=rtol, hermitian=True)

    x = jnp.dot(A_inv, b)

    return unravel(x), None


@partial_from_kwargs
def svd(A, b, *, rcond=None, x0=None):
    """
    Solve the linear system using Singular Value Decomposition.
    The diagonal shift on the matrix should be 0.

    Internally uses :func:`jax.numpy.linalg.lstsq`.

    .. note::

        If you pass only keyword arguments, this solver will directly create
        a partial capturing them.

    Args:
        A: the matrix A in Ax=b
        b: the vector b in Ax=b
        rcond: The condition number
    """
    del x0

    if not isinstance(A, jax.Array):
        A = A.to_dense()
    b, unravel = tree_ravel(b)

    x, residuals, rank, s = jnp.linalg.lstsq(A, b, rcond=rcond)

    return unravel(x), (residuals, rank, s)


@partial_from_kwargs
def cholesky(A, b, *, lower=False, x0=None):
    """
    Solve the linear system using a Cholesky Factorisation.
    The diagonal shift on the matrix should be 0.

    Internally uses :func:`jax.numpy.linalg.cho_solve`.

    .. note::

        If you pass only keyword arguments, this solver will directly create
        a partial capturing them.

    Args:
        A: the matrix A in Ax=b
        b: the vector b in Ax=b
        lower: if True uses the lower half of the A matrix
        x0: unused
    """

    del x0

    if not isinstance(A, jax.Array):
        A = A.to_dense()
    b, unravel = tree_ravel(b)

    c, low = jsp.linalg.cho_factor(A, lower=lower)
    x = jsp.linalg.cho_solve((c, low), b)
    return unravel(x), None


@partial_from_kwargs
def LU(A, b, *, trans=0, x0=None):
    """
    Solve the linear system using a LU Factorisation.
    The diagonal shift on the matrix should be 0.

    Internally uses :func:`jax.numpy.linalg.lu_solve`.

    .. note::

        If you pass only keyword arguments, this solver will directly create
        a partial capturing them.

    Args:
        A: the matrix A in Ax=b
        b: the vector b in Ax=b
        lower: if True uses the lower half of the A matrix
        x0: unused
    """

    del x0

    if not isinstance(A, jax.Array):
        A = A.to_dense()
    b, unravel = tree_ravel(b)

    lu, piv = jsp.linalg.lu_factor(A)
    x = jsp.linalg.lu_solve((lu, piv), b, trans=0)
    return unravel(x), None


# I believe this internally uses a smarter/more efficient way to
# do cholesky
@partial_from_kwargs
def solve(A, b, *, assume_a="pos", x0=None):
    """
    Solve the linear system.
    The diagonal shift on the matrix should be 0.

    Internally uses :func:`jax.numpy.solve`.

    .. note::

        If you pass only keyword arguments, this solver will directly create
        a partial capturing them.

    Args:
        A: the matrix A in Ax=b
        b: the vector b in Ax=b
        lower: if True uses the lower half of the A matrix
        x0: unused
    """
    del x0

    if not isinstance(A, jax.Array):
        A = A.to_dense()
    b, unravel = tree_ravel(b)

    x = jsp.linalg.solve(A, b, assume_a="pos")
    return unravel(x), None



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scipy import sparse as _sparse

from netket.utils.types import DType as _DType

from netket.hilbert import DiscreteHilbert as _DiscreteHilbert

from ._local_operator import LocalOperator as _LocalOperator

# export identity from here as well
from .spin import identity  # noqa: F401


def destroy(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = None
) -> _LocalOperator:
    """
    Builds the boson destruction operator :math:`\\hat{a}` acting on the `site`-th of
    the Hilbert space `hilbert`.

    If `hilbert` is a non-Bosonic space of local dimension M, it is considered
    as a bosonic space of local dimension M.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)

    D = np.array([np.sqrt(m) for m in np.arange(1, N)])
    mat = np.diag(D, 1)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)


def create(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = float
) -> _LocalOperator:
    """
    Builds the boson creation operator :math:`\\hat{a}^\\dagger` acting on the `site`-th
    of the Hilbert space `hilbert`.

    If `hilbert` is a non-Bosonic space of local dimension M, it is considered
    as a bosonic space of local dimension M.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)

    D = np.array([np.sqrt(m) for m in np.arange(1, N)])
    mat = np.diag(D, -1)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)


def number(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = float
) -> _LocalOperator:
    """
    Builds the number operator :math:`\\hat{a}^\\dagger\\hat{a}`  acting on the
    `site`-th of the Hilbert space `hilbert`.

    If `hilbert` is a non-Bosonic space of local dimension M, it is considered
    as a bosonic space of local dimension M.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)

    D = np.array([m for m in np.arange(0, N)])
    mat = np.diag(D, 0)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)


def proj(
    hilbert: _DiscreteHilbert, site: int, n: int, dtype: _DType = float
) -> _LocalOperator:
    """
    Builds the projector operator :math:`|n\\rangle\\langle n |` acting on the
    `site`-th of the Hilbert space `hilbert` and collapsing on the state with `n`
    bosons.

    If `hilbert` is a non-Bosonic space of local dimension M, it is considered
    as a bosonic space of local dimension M.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)

    if n >= N:
        raise ValueError("Cannot project on a state above the cutoff.")

    D = np.array([0 for m in np.arange(0, N)])
    D[n] = 1
    mat = np.diag(D, 0)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This file only contains deprecated bindings. To be removed.
"""

from netket.operator import fermion as _deprecated_fermion

_deprecations = {
    # June 2024, NetKet 3.13
    f"{_name}": (
        f"netket.experimental.operator.fermion.{_name} is deprecated: use "
        f"netket.operator.fermion.{_name} (netket >= 3.13)",
        getattr(_deprecated_fermion, _name),
    )
    for _name in [
        "destroy",
        "create",
        "number",
        "identity",
        "zero",
    ]
}

from netket.utils import _auto_export  # noqa: E402
from netket.utils.deprecation import (  # noqa: E402
    deprecation_getattr as _deprecation_getattr,
)

__getattr__ = _deprecation_getattr(__name__, _deprecations)
_auto_export(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings
import types
from textwrap import dedent

from netket.utils.config_flags import config
from netket.utils import module_version

_mpi4py_loaded = False
_mpi4jax_loaded = False
mpi4jax_available = False

try:
    if (
        not config.netket_mpi or config.netket_experimental_sharding  # type: ignore
    ):  # pragma: no cover  # type: ignore
        # if mpi is disabled trigger import error
        # and follow the no-mpi code path
        raise ImportError

    from mpi4py import MPI

    _mpi4py_loaded = True
    mpi4py_available = True

    # We don't use the standard communicator because Jax and
    # np/python must use different ones to avoid desync issues
    # so we insulate also from the user-space.
    MPI_py_comm = MPI.COMM_WORLD.Create(MPI.COMM_WORLD.Get_group())
    MPI_jax_comm = MPI.COMM_WORLD.Create(MPI.COMM_WORLD.Get_group())

    n_nodes = MPI_py_comm.Get_size()
    node_number = MPI_py_comm.Get_rank()
    rank = MPI_py_comm.Get_rank()

    import mpi4jax

    _mpi4jax_loaded = True
    mpi4jax_available = True

    # If we got here, mpi4jax is loaded and we are using MPI.
    # Let's check if we have multiple GPUs available. If that is the case, as netket
    # does not support multiple GPUs per rank in MPI mode, we fix that only one gpu
    # is avaialble per rank, by using local ranks.
    if config.netket_mpi_autodetect_local_gpu:  # type: ignore
        from .gpu_autorank_util import autoset_default_gpu

        autoset_default_gpu(MPI_py_comm)  # type: ignore
        del autoset_default_gpu


except ImportError:
    mpi4py_available = False
    mpi4jax_available = False
    MPI_py_comm = None  # type: ignore
    MPI_jax_comm = None  # type: ignore
    n_nodes = 1
    node_number = 0
    rank = 0

    FakeMPI = types.ModuleType("FakeMPI", "FakeMPI Module")
    FakeMPI.COMM_WORLD = None

    MPI = FakeMPI

    # Try to detect if we are running under MPI and warn that mpi4py is not installed
    if config.netket_mpi_warning and not config.netket_experimental_sharding:  # type: ignore
        _MPI_ENV_VARIABLES = [
            "OMPI_COMM_WORLD_SIZE",
            "I_MPI_HYDRA_HOST_FILE",
            "MPI_LOCALRANKID",
            "MPI_LOCALNRANKS",
        ]
        for varname in _MPI_ENV_VARIABLES:
            if varname in os.environ:
                warnings.warn(
                    dedent(
                        f"""
                    MPI WARNING: It seems you might be running Python with MPI, but
                    dependencies required by NetKet to enable MPI support are missing
                    or cannot be loaded, so MPI support is disabled.

                    NetKet will not take advantage of MPI, and every MPI rank will
                    execute the same code independently.

                    MPI dependencies are:
                      - mpi4py>=3.0.1     ....... {"available" if _mpi4py_loaded else "missing"}
                      - mpi4jax>=0.5.11   ....... {"available" if _mpi4jax_loaded else "missing"}

                    To enable MPI support, install the missing dependencies.
                    To learn more about MPI and NetKet consult the documentation at
                    https://www.netket.org/docs/getting_started.html

                    To disable this warning, set the environment variable `NETKET_MPI_WARNING=0`
                    """
                    )
                )

if mpi4py_available:
    _MIN_MPI4JAX_VERSION = (0, 3, 1)

    if module_version(mpi4jax) < _MIN_MPI4JAX_VERSION:  # type: ignore
        raise ImportError(
            f"Netket is only compatible with mpi4jax >= {_MIN_MPI4JAX_VERSION} "
            f"(you have mpi4jax == {mpi4jax.__version__}). "  # type: ignore
            "Please update it to a more recent version by running "
            "(`pip install -U mpi4jax`)."
        )


del os, warnings, dedent, config, module_version



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ["ContinuousHilbert", "Particle"]

from .continuous_hilbert import ContinuousHilbert
from .particle import Particle
from . import random
from netket.hilbert import SpinOrbitalFermions as _deprecated_SpinOrbitalFermions

_deprecations = {
    # May 2024
    "SpinOrbitalFermions": (
        "netket.experimental.hilbert.SpinOrbitalFermions is deprecated: use "
        "netket.hilbert.SpinOrbitalFermions (netket >= 3.12)",
        _deprecated_SpinOrbitalFermions,
    ),
}

from netket.utils import _auto_export
from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

__getattr__ = _deprecation_getattr(__name__, _deprecations)
_auto_export(__name__)

del _deprecation_getattr, _auto_export



# Copyright 2021 The NetKet Authors - All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Union, Protocol
from collections.abc import Callable
from collections.abc import Sequence

import optax as _optax
import jax as _jax
import numpy as _np
import flax as _flax

from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

PRNGKeyT = Any
SeedT = Union[int, PRNGKeyT]


class _SupportsDType(Protocol):
    @property
    def dtype(self) -> _np.dtype: ...


Shape = Sequence[int]
DType = Union[
    None,
    str,  # like 'float32', 'int32'
    type[Any],  # like np.float32, np.int32, float, int
    _np.dtype,  # like np.dtype('float32'), np.dtype('int32')
    _SupportsDType,  # like jnp.float32, jnp.int32
]

Array = Union[_np.ndarray, _jax.Array]
JaxArray = _jax.Array

ArrayLike = Any  # Objects that are valid inputs to (np|jnp).asarray.

NNInitFunc = Callable[[PRNGKeyT, Shape, DType], _jax.Array]

PyTree = Any
ModuleOrApplyFun = Union[_flax.linen.Module, Callable[[PyTree, Array], Array]]
ModelStateT = dict[str, PyTree]

Scalar = Any

ScalarOrSchedule = Union[Scalar, _optax.Schedule]

Optimizer = Any

_deprecations = {
    # April 2025
    "Any": (
        "netket.utils.types.Any is deprecated: use " "typing.Any (netket >= 3.17)",
        Any,
    ),
    "Union": (
        "netket.utils.types.Union is deprecated: use "
        "typing.Union or `|` (netket >= 3.17)",
        Union,
    ),
    "Protocol": (
        "netket.utils.types.Protocol is deprecated: use "
        "typing.Protocol (netket >= 3.17)",
        Protocol,
    ),
    "Callable": (
        "netket.utils.types.Callable is deprecated: use "
        "collections.abc.Callable (netket >= 3.17)",
        Callable,
    ),
    "Sequence": (
        "netket.utils.types.Sequence is deprecated: use "
        "collections.abc.Sequence (netket >= 3.17)",
        Sequence,
    ),
}

__getattr__ = _deprecation_getattr(__name__, _deprecations)
del _deprecation_getattr


del Any, Union, Protocol, Callable, Sequence



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np

import jax
from jax import numpy as jnp
from flax import linen as nn
from jax.nn.initializers import normal

from netket.utils import HashableArray
from netket.utils.types import NNInitFunc
from netket.utils.group import PermutationGroup
from netket import nn as nknn

default_kernel_init = normal(stddev=0.01)


class RBM(nn.Module):
    r"""A restricted boltzman Machine, equivalent to a 2-layer FFNN with a
    nonlinear activation function in between.
    """

    param_dtype: Any = np.float64
    """The dtype of the weights."""
    activation: Any = nknn.log_cosh
    """The nonlinear activation function."""
    alpha: float | int = 1
    """feature density. Number of features equal to alpha * input.shape[-1]"""
    use_hidden_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    use_visible_bias: bool = True
    """if True adds a bias to the input not passed through the nonlinear layer."""
    precision: Any = None
    """numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the Dense layer matrix."""
    hidden_bias_init: NNInitFunc = default_kernel_init
    """Initializer for the hidden bias."""
    visible_bias_init: NNInitFunc = default_kernel_init
    """Initializer for the visible bias."""

    @nn.compact
    def __call__(self, input):
        x = nn.Dense(
            name="Dense",
            features=int(self.alpha * input.shape[-1]),
            param_dtype=self.param_dtype,
            precision=self.precision,
            use_bias=self.use_hidden_bias,
            kernel_init=self.kernel_init,
            bias_init=self.hidden_bias_init,
        )(input)
        x = self.activation(x)
        x = jnp.sum(x, axis=-1)

        if self.use_visible_bias:
            v_bias = self.param(
                "visible_bias",
                self.visible_bias_init,
                (input.shape[-1],),
                self.param_dtype,
            )
            out_bias = jnp.dot(input, v_bias)
            return x + out_bias
        else:
            return x


class RBMModPhase(nn.Module):
    r"""
    A fully connected Restricted Boltzmann Machine (RBM) with real-valued parameters.

    In this case, two RBMs are taken to parametrize, respectively, the real
    and imaginary part of the log-wave-function, as introduced in Torlai et al.,
    Nature Physics 14, 447–450(2018).

    This type of RBM has spin 1/2 hidden units and is defined by:

    .. math::

        \Psi(s_1,\dots s_N) = e^{\sum_i^N a_i s_i} \times \Pi_{j=1}^M
            \cosh \left(\sum_i^N W_{ij} s_i + b_j \right)

    for arbitrary local quantum numbers :math:`s_i`.
    """

    param_dtype: Any = np.float64
    """The dtype of the weights."""
    activation: Any = nknn.log_cosh
    """The nonlinear activation function."""
    alpha: float | int = 1
    """feature density. Number of features equal to alpha * input.shape[-1]"""
    use_hidden_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    precision: Any = None
    """Numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the Dense layer matrix."""
    hidden_bias_init: NNInitFunc = default_kernel_init
    """Initializer for the hidden bias."""

    @nn.compact
    def __call__(self, x):
        re = nn.Dense(
            features=int(self.alpha * x.shape[-1]),
            param_dtype=self.param_dtype,
            use_bias=self.use_hidden_bias,
            precision=self.precision,
            kernel_init=self.kernel_init,
            bias_init=self.hidden_bias_init,
        )(x)
        re = self.activation(re)
        re = jnp.sum(re, axis=-1)

        im = nn.Dense(
            features=int(self.alpha * x.shape[-1]),
            param_dtype=self.param_dtype,
            use_bias=self.use_hidden_bias,
            precision=self.precision,
            kernel_init=self.kernel_init,
            bias_init=self.hidden_bias_init,
        )(x)
        im = self.activation(im)
        im = jnp.sum(im, axis=-1)

        return re + 1j * im


class RBMMultiVal(nn.Module):
    """
    A fully connected Restricted Boltzmann Machine (see :class:`netket.models.RBM`) suitable for large local hilbert spaces.
    Local quantum numbers are passed through a one hot encoding that maps them onto
    an enlarged space of +/- 1 spins. In turn, these quantum numbers are used with a
    standard :class:`~netket.models.RBM` wave function.
    """

    n_classes: int
    """The number of classes in the one-hot encoding"""
    param_dtype: Any = np.float64
    """The dtype of the weights."""
    activation: Any = nknn.log_cosh
    """The nonlinear activation function."""
    alpha: float | int = 1
    """feature density. Number of features equal to alpha * input.shape[-1]"""
    use_hidden_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    use_visible_bias: bool = True
    """if True adds a bias to the input not passed through the nonlinear layer."""
    precision: Any = None
    """Numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the Dense layer matrix."""
    hidden_bias_init: NNInitFunc = default_kernel_init
    """Initializer for the hidden bias."""
    visible_bias_init: NNInitFunc = default_kernel_init
    """Initializer for the visible bias."""

    def setup(self):
        self.RBM = RBM(
            param_dtype=self.param_dtype,
            activation=self.activation,
            alpha=self.alpha,
            use_hidden_bias=self.use_hidden_bias,
            use_visible_bias=self.use_visible_bias,
            kernel_init=self.kernel_init,
            hidden_bias_init=self.hidden_bias_init,
            visible_bias_init=self.visible_bias_init,
        )

    def __call__(self, x):
        batches = x.shape[:-1]
        N = x.shape[-1]

        # do the one hot encoding: output x.shape +(n_classes,)
        x_oh = jax.nn.one_hot(x, self.n_classes)
        # vectorize the last two dimensions
        x_oh = jnp.reshape(
            x_oh,
            (
                *batches,
                self.n_classes * N,
            ),
        )
        # apply the rbm to this output
        return self.RBM(x_oh)


class RBMSymm(nn.Module):
    """A symmetrized RBM using the :func:`netket.nn.DenseSymm` layer internally."""

    symmetries: HashableArray | PermutationGroup
    """A group of symmetry operations (or array of permutation indices) over which the layer should be invariant.
    Numpy/Jax arrays must be wrapped into an :class:`netket.utils.HashableArray`.
    """
    param_dtype: Any = np.float64
    """The dtype of the weights."""
    activation: Any = nknn.log_cosh
    """The nonlinear activation function."""
    alpha: float | int = 1
    """feature density. Number of features equal to alpha * input.shape[-1]"""
    use_hidden_bias: bool = True
    """if True uses a bias in the dense layer (hidden layer bias)."""
    use_visible_bias: bool = True
    """if True adds a bias to the input not passed through the nonlinear layer."""
    precision: Any = None
    """Numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    kernel_init: NNInitFunc = normal(stddev=0.1)
    """Initializer for the Dense layer matrix."""
    hidden_bias_init: NNInitFunc = normal(stddev=0.1)
    """Initializer for the hidden bias."""
    visible_bias_init: NNInitFunc = normal(stddev=0.1)
    """Initializer for the visible bias."""

    def setup(self):
        self.n_symm, self.n_sites = np.asarray(self.symmetries).shape
        self.features = int(self.alpha * self.n_sites / self.n_symm)
        if self.alpha > 0 and self.features == 0:
            raise ValueError(
                f"RBMSymm: alpha={self.alpha} is too small "
                f"for {self.n_symm} permutations, alpha ≥ {self.n_symm / self.n_sites} is needed."
            )

    @nn.compact
    def __call__(self, x_in):
        x = x_in
        if x.ndim < 3:
            x = jnp.expand_dims(x, -2)
        x = nknn.DenseSymm(
            name="Dense",
            mode="matrix",
            symmetries=self.symmetries,
            features=self.features,
            param_dtype=self.param_dtype,
            use_bias=self.use_hidden_bias,
            kernel_init=self.kernel_init,
            bias_init=self.hidden_bias_init,
            precision=self.precision,
        )(x)
        x = self.activation(x)

        x = x.reshape(-1, self.features * self.n_symm)
        x = jnp.sum(x, axis=-1)

        if self.use_visible_bias:
            v_bias = self.param(
                "visible_bias", self.visible_bias_init, (1,), self.param_dtype
            )
            out_bias = v_bias[0] * jnp.sum(x_in, axis=-1)
            return x + out_bias
        else:
            return x



# Copyright 2021 The NetKet Authors - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp

import netket.jax as nkjax
from netket.utils.types import Array, PyTree


LimitsDType = tuple[float | None, float | None]
"""Type of the dt limits field, having independently optional upper and lower bounds."""


def expand_dim(tree: PyTree, sz: int) -> PyTree:
    """
    creates a new pytree with same structure as input `tree`, but where very leaf
    has an extra dimension at 0 with size `sz`.
    """

    def _expand(x):
        return jnp.zeros((sz, *x.shape), dtype=x.dtype)

    return jax.tree_util.tree_map(_expand, tree)


def propose_time_step(
    dt: float, scaled_error: float, error_order: int, limits: LimitsDType
) -> float:
    """
    Propose an updated dt based on the scheme suggested in Numerical Recipes, 3rd ed.
    """
    SAFETY_FACTOR = 0.95
    err_exponent = -1.0 / (1 + error_order)
    return jnp.clip(
        dt * SAFETY_FACTOR * scaled_error**err_exponent,
        limits[0],
        limits[1],
    )


def set_flag_jax(condition, flags, flag):
    """
    If `condition` is true, `flags` is updated by setting `flag` to 1.
    This is equivalent to the following code, but compatible with jax.jit:
        if condition:
            flags |= flag
    """
    return jax.lax.cond(
        condition,
        lambda x: x | flag,
        lambda x: x,
        flags,
    )


def scaled_error(y, y_err, atol, rtol, *, last_norm_y=None, norm_fn) -> float:
    norm_y = norm_fn(y)
    scale = (atol + jnp.maximum(norm_y, last_norm_y) * rtol) / nkjax.tree_size(y_err)
    return norm_fn(y_err) / scale, norm_y


def euclidean_norm(x: PyTree | Array) -> float:
    """
    Computes the Euclidean L2 norm of the Array or PyTree intended as a flattened array
    """
    if isinstance(x, jnp.ndarray):
        return jnp.sqrt(jnp.sum(jnp.abs(x) ** 2))
    else:
        return jnp.sqrt(
            jax.tree_util.tree_reduce(
                lambda x, y: x + y,
                jax.tree_util.tree_map(lambda x: jnp.sum(jnp.abs(x) ** 2), x),
            )
        )


def maximum_norm(x: PyTree | Array) -> float:
    """
    Computes the maximum norm of the Array or PyTree intended as a flattened array
    """
    if isinstance(x, jnp.ndarray):
        return jnp.max(jnp.abs(x))
    else:
        return jnp.sqrt(
            jax.tree_util.tree_reduce(
                jnp.maximum,
                jax.tree_util.tree_map(lambda x: jnp.max(jnp.abs(x)), x),
            )
        )



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from flax import linen as nn
from flax.linen.dtypes import promote_dtype
from jax import numpy as jnp
from jax.nn.initializers import orthogonal, zeros

from netket.utils.types import DType, NNInitFunc


default_kernel_init = orthogonal()


class RNNCell(nn.Module):
    """Recurrent neural network cell that updates the hidden memory at each site."""

    features: int
    """output feature density, should be the last dimension."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""

    @abc.abstractmethod
    def __call__(self, inputs, cell_mem, hidden):
        """
        Applies the RNN cell to a batch of input sites at a given index.

        Args:
            inputs: input data with dimensions (batch, in_features).
            cell_mem: cell memory from the previous site with dimensions (batch, features).
            hidden: hidden memories from the previous neighbors with dimensions (batch, n_neighbors, features).

        Returns:
             * :code:`cell_mem`
                the updated cell memory with dimensions :code:`(batch, self.features)`.
             * :code:`outputs`
                the updated hidden memory with dimensions :code:`(batch, self.features)`,
                also serves as the output data at the current site for the
                :class:`netket.experimental.nn.rnn.RNNLayer` layer.
        """


class LSTMCell(RNNCell):
    """Long short-term memory cell."""

    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weight matrix."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    @nn.compact
    def __call__(self, inputs, cell_mem, hidden):
        batch_size = inputs.shape[0]
        in_features = inputs.shape[-1]

        hidden = hidden.reshape((batch_size, -1))
        hid_features = hidden.shape[-1]
        in_cat = jnp.concatenate([inputs, hidden], axis=-1)

        kernel = self.param(
            "kernel",
            self.kernel_init,
            (in_features + hid_features, self.features * 4),
            self.param_dtype,
        )
        bias = self.param(
            "bias",
            self.bias_init,
            (self.features * 4,),
            self.param_dtype,
        )

        in_cat, kernel, bias = promote_dtype(in_cat, kernel, bias, dtype=None)

        ifgo = nn.sigmoid(in_cat @ kernel + bias)
        i, f, g, o = jnp.split(ifgo, 4, axis=-1)

        # sigmoid -> tanh
        g = g * 2 - 1

        cell_mem = f * cell_mem + i * g
        outputs = o * nn.tanh(cell_mem)
        return cell_mem, outputs


class GRU1DCell(RNNCell):
    """Gated recurrent unit cell.

    Only supports one previous neighbor at each site.
    """

    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weight matrix."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    # cell_mem is not used
    @nn.compact
    def __call__(self, inputs, cell_mem, hidden):
        batch_size = inputs.shape[0]
        in_features = inputs.shape[-1]

        hidden = hidden.reshape((batch_size, -1))
        hid_features = hidden.shape[-1]
        in_cat = jnp.concatenate([inputs, hidden], axis=-1)

        rz_kernel = self.param(
            "rz_kernel",
            self.kernel_init,
            (in_features + hid_features, self.features * 2),
            self.param_dtype,
        )
        rz_bias = self.param(
            "rz_bias",
            self.bias_init,
            (self.features * 2,),
            self.param_dtype,
        )
        n_kernel = self.param(
            "n_kernel",
            self.kernel_init,
            (in_features + hid_features, self.features),
            self.param_dtype,
        )
        n_bias = self.param(
            "n_bias",
            self.bias_init,
            (self.features),
            self.param_dtype,
        )

        in_cat, rz_kernel, rz_bias, n_kernel, n_bias = promote_dtype(
            in_cat, rz_kernel, rz_bias, n_kernel, n_bias, dtype=None
        )

        rz = nn.sigmoid(in_cat @ rz_kernel + rz_bias)
        r, z = jnp.split(rz, 2, axis=-1)

        in_cat = jnp.concatenate([inputs, r * hidden], axis=-1)
        n = nn.tanh(in_cat @ n_kernel + n_bias)

        outputs = (1 - z) * n + z * hidden
        return cell_mem, outputs



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains functions that act on hilbert spaces necessary
for sampling.

If you define custom hilbert spaces, and want to sample from it, you
should read carefully.

This module *exports* two functions:
 - `random_state(hilbert, key, batch_size, dtype)` which generates a batch of random
 states in the given hilbert space, in an array with the specified dtype.
 - `flip_state(hilb, key, states, indices)` which, for every state σ in the batch of
 states, considers σᵢ and returns a new state where that entry is different from the
 previous. The new configuration is selected with uniform probability among the local
 possible configurations.

Hilbert spaces must at least implement a `random_state` to support sampling.
`flip_state` is only necessary in order to use LocalRule samplers.
All hilbert spaces in netket implement both.

How to implement the two functions above
----------------------------------------

While the *exported* function acts on batches, sometimes it is hard to implement
the function on batches. Therefore they can be implemented either as a scalar
function, that gets jax.vmap-ed automatically, or directly as a batched rule.
Of course, if you implement the batched rule you will most likely see better
performance.

In order to implement the scalar rule for your custom hilbert object `MyHilbert`
you should define a function taking 4 inputs, the hilbert space, a jax PRNG key and
the dtype of the desired result. For random operations you should use the key
provided.

@netket.utils.dispatch.dispatch
def random_state(hilb: MyHilbert, key, dtype):
    return mystate

@netket.utils.dispatch.dispatch
def random_state(hilb: MyHilbert, key, batches: int, dtype):
    return mystate

flip_state is implemented in the same way, through

@netket.utils.dispatch.dispatch
def flip_state_scalar(hilb: Fock, key, σ, idx):
    return new_state, oldval

There is a vmapped default fallback for the batched version.

@netket.utils.dispatch.dispatch
def flip_state_batch(hilb: Fock, key, σ, idx):
    return new_states, oldvals

"""

from netket.utils import _hide_submodules

from . import particle
from netket.hilbert.random.base import flip_state, random_state

_hide_submodules(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING
from flax import serialization


from ..version_check import module_version

if TYPE_CHECKING:
    from flax import nnx


# Flax version 0.10.0 and later have a bug in nnx.to_linen()
# And cannot serialize some NodeDef Mappings that they put in the model_state.
if module_version("flax") >= (0, 10, 0):
    from flax import nnx

    try:

        def serialize_flat_mapping(NodeDef):
            return {}

        def deserialize_flat_mapping(NodeDef, _):
            return NodeDef

        serialization.register_serialization_state(
            nnx.graph.NodeDef,
            serialize_flat_mapping,
            deserialize_flat_mapping,
        )
    except Exception:
        pass

    if module_version("flax") < (0, 10, 2):
        from flax import nnx
        from flax.core import FrozenDict
        from flax.nnx.bridge import ToLinen

        def to_linen(nnx_class, *args, name: str | None = None, **kwargs):
            """Shortcut of `nnx.bridge.ToLinen` if user is not changing any of its default fields."""
            return ToLinen(nnx_class, args=args, kwargs=FrozenDict(kwargs), name=name)

        setattr(nnx.bridge, "to_linen", to_linen)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._discrete_operator import DiscreteOperator


class SpecialHamiltonian(DiscreteOperator):
    def to_local_operator(self):
        raise NotImplementedError(
            f"Must implemented to_local_operator for {type(self)}"
        )

    def conjugate(self, *, concrete: bool = True):
        return self.to_local_operator().conjugate(concrete=concrete)

    def __add__(self, other):
        if type(self) is type(other):
            res = self.copy()
            res = res.__iadd__(other)
            if res is not NotImplemented:
                return res

        return self.to_local_operator() + other

    def __sub__(self, other):
        if type(self) is type(other):
            res = self.copy()
            res = res.__isub__(other)
            if res is not NotImplemented:
                return res

        return self.to_local_operator() - other

    def __radd__(self, other):
        return self.to_local_operator().__radd__(other)

    def __rsub__(self, other):
        return self.to_local_operator().__rsub__(other)

    def __iadd__(self, other):
        if type(self) is type(other):
            res = self._iadd_same_hamiltonian(other)
            return res

        return NotImplemented

    def __isub__(self, other):
        if type(self) is type(other):
            res = self._isub_same_hamiltonian(other)
            return res

        return NotImplemented

    def __mul__(self, other):
        return self.to_local_operator().__mul__(other)

    def __rmul__(self, other):
        return self.to_local_operator().__rmul__(other)

    def _op__matmul__(self, other):
        if hasattr(other, "to_local_operator"):
            other = other.to_local_operator()
        return self.to_local_operator().__matmul__(other)

    def __neg__(self):
        return -1 * self.to_local_operator()

    def _op__rmatmul__(self, other):
        if hasattr(other, "to_local_operator"):
            other = other.to_local_operator()

        return self.to_local_operator().__matmul__(other)



import jax
import jax.numpy as jnp


def _remove_prngkey(val):
    if isinstance(val, jax.Array) and jnp.issubdtype(val.dtype, jax.dtypes.prng_key):
        return jax.random.key_data(val)
    else:
        return val


def _restore_prngkey(target_val, state_dict_val):
    if isinstance(target_val, jax.Array) and jnp.issubdtype(
        target_val.dtype, jax.dtypes.prng_key
    ):
        return jax.random.wrap_key_data(
            state_dict_val, impl=jax.random.key_impl(target_val)
        )
    else:
        return state_dict_val


def remove_prngkeys(tree):
    """
    Removes PRNG keys from a pytree, replacing them with their key_data.

    This is useful when serializing a pytree to disk, as PRNG keys are not
    serializable.

    Args:
        tree: the pytree to remove PRNG keys from.

    Returns:
        The pytree with PRNG keys replaced by their key_data.
    """
    return jax.tree.map(_remove_prngkey, tree)


def restore_prngkeys(target, state_dict):
    """
    Restores PRNG keys in a pytree from their key_data, reversing the operation
    performed by `remove_prngkeys`.

    Args:
        target: the target pytree to restore PRNG keys in.
        state_dict: the pytree containing the key_data to restore.
    """
    return jax.tree.map(_restore_prngkey, target, state_dict)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from collections.abc import Callable
from collections.abc import Sequence
import operator

import jax
import jax.numpy as jnp
from jax.util import safe_map

from ._utils_tree import tree_leaf_iscomplex, eval_shape

map = safe_map


def _ensure_index(x: Any) -> int | tuple[int, ...]:
    """Ensure x is either an index or a tuple of indices."""
    try:
        return operator.index(x)
    except TypeError:
        return tuple(map(operator.index, x))


def grad(
    fun: Callable,
    argnums: int | Sequence[int] = 0,
    has_aux: bool = False,
    allow_int: bool = False,
) -> Callable:
    """Creates a function which evaluates the gradient of ``fun``.

    Args:
      fun: Function to be differentiated. Its arguments at positions specified by
        ``argnums`` should be arrays, scalars, or standard Python containers.
        Argument arrays in the positions specified by ``argnums`` must be of
        inexact (i.e., floating-point or complex) type. It
        should return a scalar (which includes arrays with shape ``()`` but not
        arrays with shape ``(1,)`` etc.)
      argnums: Optional, integer or sequence of integers. Specifies which
        positional argument(s) to differentiate with respect to (default 0).
      has_aux: Optional, bool. Indicates whether ``fun`` returns a pair where the
        first element is considered the output of the mathematical function to be
        differentiated and the second element is auxiliary data. Default False.
      holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
        holomorphic. If True, inputs and outputs must be complex. Default False.
      allow_int: Optional, bool. Whether to allow differentiating with
        respect to integer valued inputs. The gradient of an integer input will
        have a trivial vector-space dtype (float0). Default False.

    Returns:
      A function with the same arguments as ``fun``, that evaluates the gradient
      of ``fun``. If ``argnums`` is an integer then the gradient has the same
      shape and type as the positional argument indicated by that integer. If
      argnums is a tuple of integers, the gradient is a tuple of values with the
      same shapes and types as the corresponding arguments. If ``has_aux`` is True
      then a pair of (gradient, auxiliary_data) is returned.

    For example:

    >>> import jax
    >>>
    >>> grad_tanh = jax.grad(jax.numpy.tanh)
    >>> print(grad_tanh(0.2))
    0.9610429829661166
    """
    value_and_grad_f = value_and_grad(
        fun, argnums, has_aux=has_aux, allow_int=allow_int
    )

    def grad_f(*args, **kwargs):
        _, g = value_and_grad_f(*args, **kwargs)
        return g

    def grad_f_aux(*args, **kwargs):
        (_, aux), g = value_and_grad_f(*args, **kwargs)
        return g, aux

    return grad_f_aux if has_aux else grad_f


def value_and_grad(
    fun: Callable,
    argnums: int | Sequence[int] = 0,
    has_aux: bool = False,
    allow_int: bool = False,
) -> Callable[..., tuple[Any, Any]]:
    """Create a function which evaluates both ``fun`` and the gradient of ``fun``.

    Args:
      fun: Function to be differentiated. Its arguments at positions specified by
        ``argnums`` should be arrays, scalars, or standard Python containers. It
        should return a scalar (which includes arrays with shape ``()`` but not
        arrays with shape ``(1,)`` etc.)
      argnums: Optional, integer or sequence of integers. Specifies which
        positional argument(s) to differentiate with respect to (default 0).
      has_aux: Optional, bool. Indicates whether ``fun`` returns a pair where the
        first element is considered the output of the mathematical function to be
        differentiated and the second element is auxiliary data. Default False.
      allow_int: Optional, bool. Whether to allow differentiating with
        respect to integer valued inputs. The gradient of an integer input will
        have a trivial vector-space dtype (float0). Default False.

    Returns:
      A function with the same arguments as ``fun`` that evaluates both ``fun``
      and the gradient of ``fun`` and returns them as a pair (a two-element
      tuple). If ``argnums`` is an integer then the gradient has the same shape
      and type as the positional argument indicated by that integer. If argnums is
      a sequence of integers, the gradient is a tuple of values with the same
      shapes and types as the corresponding arguments.
    """

    argnums = _ensure_index(argnums)
    docstr = (
        "Value and gradient of {fun} with respect to positional "
        "argument(s) {argnums}. Takes the same arguments as {fun} but "
        "returns a two-element tuple where the first element is the value "
        "of {fun} and the second element is the gradient, which has the "
        "same shape as the arguments at positions {argnums}."
    )
    docstr = docstr.format(fun=fun, argnums=argnums)

    # @wraps(fun, docstr=docstr)
    def value_and_grad_f(*args, **kwargs):
        out_shape = eval_shape(fun, *args, has_aux=has_aux, **kwargs)

        _args_iterable = (argnums,) if isinstance(argnums, int) else argnums

        # only check if derivable arguments are complex
        if tree_leaf_iscomplex([args[i] for i in _args_iterable]):
            if jnp.iscomplexobj(out_shape):  # C -> C
                return jax.value_and_grad(
                    fun,
                    argnums=argnums,
                    has_aux=has_aux,
                    allow_int=allow_int,
                    holomorphic=True,
                )(*args, **kwargs)
            else:  # C -> R
                raise RuntimeError("C->R function detected, but not supported.")
        else:
            if jnp.iscomplexobj(out_shape):  # R -> C

                def grad_rc(*args, **kwargs):
                    if has_aux:

                        def real_fun(*args, **kwargs):
                            val, aux = fun(*args, **kwargs)
                            return val.real, aux

                        def imag_fun(*args, **kwargs):
                            val, aux = fun(*args, **kwargs)
                            return val.imag, aux

                        out_r, grad_r, aux = jax.value_and_grad(
                            real_fun, argnums=argnums, has_aux=True, allow_int=allow_int
                        )(*args, **kwargs)
                        out_j, grad_j, _ = jax.value_and_grad(
                            imag_fun, argnums=argnums, has_aux=True, allow_int=allow_int
                        )(*args, **kwargs)

                    else:
                        real_fun = lambda *args, **kwargs: fun(*args, **kwargs).real
                        imag_fun = lambda *args, **kwargs: fun(*args, **kwargs).imag

                        out_r, grad_r = jax.value_and_grad(
                            real_fun,
                            argnums=argnums,
                            has_aux=False,
                            allow_int=allow_int,
                        )(*args, **kwargs)
                        out_j, grad_j = jax.value_and_grad(
                            imag_fun,
                            argnums=argnums,
                            has_aux=False,
                            allow_int=allow_int,
                        )(*args, **kwargs)

                    out = out_r + 1j * out_j
                    grad = jax.tree_util.tree_map(
                        lambda re, im: re + 1j * im, grad_r, grad_j
                    )

                    if has_aux:
                        return out, grad, aux
                    else:
                        return out, grad

                return grad_rc(*args, **kwargs)
            else:  # R -> R
                return jax.value_and_grad(
                    fun, argnums=argnums, has_aux=has_aux, allow_int=allow_int
                )(*args, **kwargs)

    return value_and_grad_f



# Copyright 2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.Mon


from flax import linen as nn
import jax
from jax import numpy as jnp

from jax.nn.initializers import normal

from netket.hilbert import HomogeneousHilbert
from netket.utils.types import NNInitFunc
from netket.jax import dtype_complex
from netket.utils.types import DType

default_kernel_init = normal(stddev=0.01)


class MPDOPeriodic(nn.Module):
    r"""
    A Matrix Product Density Operator (MPDO) with periodic boundary conditions for a quantum mixed state of discrete
    degrees of freedom. The purification is used.

    The MPDO is defined as (see `F. Verstraete, J. J. García-Ripoll, J. I. Cirac, Phys. Rev. Lett. 93, 207204 (2004) <https://doi.org/10.1103/PhysRevLett.93.207204>`_).

    .. math::
        \rho(s_1,\dots, s_N, s_1',\dots, s_N') = \sum_{\alpha_1, \dots, \alpha_{N-1}} \mathrm{Tr} \left[ M^{\alpha_1}_{s_1,s_1'} \dots M^{\alpha_{N-1}}_{s_N, s_N'} M^{\alpha_N}_{s_1, s_1'} \right],

    for arbitrary local quantum numbers :math:`s_i` and :math:`s_i'`, where :math:`M^{\alpha_i}_{s_i,s_i'}` are :math:`D^2 \times D^2` matrices that can be decomposed as

    .. math::
        M^{\alpha_i}_{s_i,s_i'} = \sum_{a=1}^{\chi} A^{\alpha_i, a}_{s_i} \otimes (A^{\alpha_i, a}_{s_i'})^*,

    with :math:`A^{\alpha_i, a}_{s_i}` being :math:`D \times D` matrices. The bond dimension is denoted by :math:`D` and the Kraus dimension by :math:`\chi`, which corresponds to the variable `kraus_dim` in the code.

    The periodic boundary conditions imply that there are connections between the first and the last tensors, forming a trace over the product of matrices for the entire system.

    The implementation is based on `this paper <https://arxiv.org/abs/2401.14243>`_.
    """

    hilbert: HomogeneousHilbert
    """Hilbert space on which the state is defined."""
    bond_dim: int
    """Bond dimension of the MPDO tensors. See formula above."""
    kraus_dim: int = 2
    """The local Kraus dimension of the MPDO tensors. See formula above."""
    symperiod: bool | None = None
    """
    Periodicity in the chain of MPDO tensors.
    The chain of MPDO tensors is constructed as a sequence of identical
    unit cells consisting of symperiod tensors. if None, symperiod equals the
    number of physical degrees of freedom.
    """
    unroll: int = 1
    """the number of scan iterations to unroll within a single iteration of a loop."""
    checkpoint: bool = True
    """Whether to use jax.checkpoint on the scan function for memory efficiency."""
    kernel_init: NNInitFunc = default_kernel_init
    """the initializer for the MPS weights. This is added to an identity tensor."""
    param_dtype: DType = float
    """complex or float, whether the variational parameters of the MPDO are real or complex."""

    def setup(self):
        L, d, D, Χ = (
            self.hilbert.size,
            self.hilbert.local_size,
            self.bond_dim,
            self.kraus_dim,
        )
        self._L, self._d, self._D, self._Χ = L, d, D, Χ

        # determine the shape of the unit cell
        if self.symperiod is None:
            self._symperiod = L
        else:
            self._symperiod = self.symperiod

        if L % self._symperiod == 0 and self._symperiod > 0:
            unit_cell_shape = (self._symperiod, d, D, D, Χ)
        else:
            raise AssertionError(
                "The number of degrees of freedom of the Hilbert space needs to be a multiple of the period of the MPS"
            )

        iden_tensors = jnp.repeat(
            jnp.eye(self.bond_dim, dtype=self.param_dtype)[jnp.newaxis, :, :],
            self._symperiod * d * Χ,
            axis=0,
        )
        iden_tensors = iden_tensors.reshape(self._symperiod, d, D, D, Χ)
        self.tensors = (
            self.param("tensors", self.kernel_init, unit_cell_shape, self.param_dtype)
            + iden_tensors
        )

    def __call__(self, x):
        """
        Queries this MPDO for the input configurations x, which should contain
        rows and columns entries concatenated.

        Args:
            x: the input configuration
        """
        x = jnp.atleast_2d(x)
        assert x.shape[-1] == 2 * self.hilbert.size
        qn_x = self.hilbert.states_to_local_indices(x)

        ρ = jax.vmap(self.contract_mpdo, in_axes=0)(qn_x)
        return jnp.log(ρ.astype(dtype_complex(self.param_dtype)))

    def contract_mpdo(self, qn):
        """
        Internal function, used to contract the tensor network with some input
        tensor.

        Args:
            qn: The input tensor to be contracted with this MPDO
        """
        # create all tensors in mps from unit cell
        all_tensors = jnp.tile(self.tensors, (self._L // self._symperiod, 1, 1, 1, 1))

        edge = jnp.eye(self._D**2, dtype=self.param_dtype)

        def base_scan_func(edge, pair):
            tensor, qn_r, qn_c = pair
            tensor_contracted = jnp.einsum(
                "ijk,lmk->iljm", tensor[qn_r, :], jnp.conj(tensor[qn_c, :])
            )
            matrix = tensor_contracted.reshape(self._D**2, self._D**2)
            edge = edge @ matrix
            return edge, None

        # Apply jax.checkpoint conditionally to the base_scan_func
        scan_func = (
            jax.checkpoint(base_scan_func) if self.checkpoint else base_scan_func
        )

        qn_r, qn_c = jnp.split(qn, 2, axis=-1)
        edge, _ = jax.lax.scan(
            scan_func, edge, (all_tensors, qn_r, qn_c), unroll=self.unroll
        )
        rho = jnp.trace(edge)
        return rho


class MPDOOpen(nn.Module):
    r"""
    A Matrix Product Density Operator (MPDO) with open boundary conditions for a quantum mixed state of discrete
    degrees of freedom. The purification is used.

    The MPDO is defined as (see `F. Verstraete, J. J. García-Ripoll, J. I. Cirac, Phys. Rev. Lett. 93, 207204 (2004) <https://doi.org/10.1103/PhysRevLett.93.207204>`_).

    .. math::
        \rho(s_1,\dots, s_N, s_1',\dots, s_N') = \sum_{\alpha_1, \dots, \alpha_{N-1}} \mathrm{Tr} \left[ M^{\alpha_1}_{s_1,s_1'} \dots M^{\alpha_{N-1}}_{s_N, s_N'} \right],

    for arbitrary local quantum numbers :math:`s_i` and :math:`s_i'`, where :math:`M^{\alpha_i}_{s_i,s_i'}` are :math:`D^2 \times D^2` matrices for :math:`i=2, \dots, N-1`, and vectors for :math:`i=1, N`, that can be decomposed as

    .. math::
        M^{\alpha_i}_{s_i,s_i'} = \sum_{a=1}^{\chi} A^{\alpha_i, a}_{s_i} \otimes (A^{\alpha_i, a}_{s_i'})^*,

    with :math:`A^{\alpha_i, a}_{s_i}` being :math:`D \times D` matrices for the bulk of the chain (:math:`i=2, \dots, N-1`) and :math:`D`-dimensional vectors for the edges (:math:`i=1, N`). The bond dimension is denoted by :math:`D` and the Kraus dimension by :math:`\chi`, which corresponds to the variable `kraus_dim` in the code.

    The open boundary conditions imply that there are no connections between the first and the last tensors in the trace.

    The implementation is based on `this paper <https://arxiv.org/abs/2401.14243>`_.
    """

    hilbert: HomogeneousHilbert
    """Hilbert space on which the state is defined."""
    bond_dim: int
    """Bond dimension of the MPDO tensors. See formula above."""
    kraus_dim: int = 2
    """The local Kraus dimension of the MPDO tensors. See formula above."""
    unroll: int = 1
    """the number of scan iterations to unroll within a single iteration of a loop."""
    checkpoint: bool = True
    """Whether to use jax.checkpoint on the scan function for memory efficiency."""
    kernel_init: NNInitFunc = default_kernel_init
    """the initializer for the MPS weights. This is added to an identity tensor."""
    param_dtype: DType = float
    """complex or float, whether the variational parameters of the MPDO are real or complex."""

    def setup(self):
        L, d, D, Χ = (
            self.hilbert.size,
            self.hilbert.local_size,
            self.bond_dim,
            self.kraus_dim,
        )
        self._L, self._d, self._D, self._Χ = L, d, D, Χ
        self.param_dtype_cplx = dtype_complex(self.param_dtype)

        iden_boundary_tensor = jnp.ones((d, D, Χ), dtype=self.param_dtype)
        self.left_tensors = (
            self.param("left_tensors", self.kernel_init, (d, D, Χ), self.param_dtype)
            + iden_boundary_tensor
        )
        self.right_tensors = (
            self.param("right_tensors", self.kernel_init, (d, D, Χ), self.param_dtype)
            + iden_boundary_tensor
        )

        # determine the shape of the unit cell
        unit_cell_shape = (L - 2, d, D, D, Χ)

        # define diagonal tensors with correct unit cell shape
        iden_tensors = jnp.repeat(
            jnp.eye(D, dtype=self.param_dtype)[jnp.newaxis, :, :],
            (L - 2) * d * Χ,
            axis=0,
        )
        iden_tensors = iden_tensors.reshape(L - 2, d, D, D, Χ)
        self.middle_tensors = (
            self.param(
                "middle_tensors", self.kernel_init, unit_cell_shape, self.param_dtype
            )
            + iden_tensors
        )

    def __call__(self, x):
        """
        Queries this MPDO for the input configurations x, which should contain
        rows and columns entries concatenated.

        Args:
            x: the input configuration
        """
        x = jnp.atleast_2d(x)
        assert x.shape[-1] == 2 * self.hilbert.size
        qn_x = self.hilbert.states_to_local_indices(x)

        ρ = jax.vmap(self.contract_mpdo)(qn_x)

        return jnp.log(ρ.astype(dtype_complex(self.param_dtype)))

    def contract_mpdo(self, qn):
        """
        Internal function, used to contract the tensor network with some input
        tensor.

        Args:
            qn: The input tensor to be contracted with this MPDO
        """
        qn_r, qn_c = jnp.split(qn, 2, axis=-1)
        left_edge = (
            self.left_tensors[qn_r[0], :] @ jnp.conj(self.left_tensors[qn_c[0], :]).T
        )
        right_edge = (
            self.right_tensors[qn_r[-1], :]
            @ jnp.conj(self.right_tensors[qn_c[-1], :]).T
        )

        def base_scan_func(edge, pair):
            tensor, qn_r, qn_c = pair
            triangle_tensor = jnp.einsum("ijk,il->jkl", tensor[qn_r, :], edge)
            edge = jnp.einsum("jkl,lmk->jm", triangle_tensor, jnp.conj(tensor[qn_c, :]))
            return edge, None

        # Apply jax.checkpoint conditionally to the base_scan_func
        scan_func = (
            jax.checkpoint(base_scan_func) if self.checkpoint else base_scan_func
        )

        edge, _ = jax.lax.scan(
            scan_func,
            left_edge,
            (self.middle_tensors, qn_r[1:-1], qn_c[1:-1]),
            unroll=self.unroll,
        )

        rho = jnp.einsum("ij,ij->", edge, right_edge)
        return rho



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from typing_extensions import Self

import numpy as np

from netket.utils import StaticRange

from .homogeneous import HomogeneousHilbert


class Qubit(HomogeneousHilbert):
    r"""Hilbert space obtained as tensor product of local qubit states."""

    def __init__(self, N: int = 1):
        r"""Initializes a qubit hilbert space.

        Args:
            N: Number of qubits.

        Examples:
            Simple spin hilbert space.

            >>> from netket.hilbert import Qubit
            >>> hi = Qubit(N=100)
            >>> print(hi.size)
            100
        """
        super().__init__(StaticRange(0, 1, 2, dtype=np.int8), N)  # type: ignore[arg-type]

    def __pow__(self, n):
        return Qubit(self.size * n)

    def _mul_sametype_(self, other):
        assert type(self) == type(other)
        return Qubit(self.size + other.size)

    def ptrace(self, sites: int | Sequence[int]) -> Self | None:
        if isinstance(sites, int):
            sites = [sites]

        for site in sites:
            if site < 0 or site >= self.size:
                raise ValueError(
                    f"Site {site} not in this hilbert space of site {self.size}"
                )

        Nsites = len(sites)

        if self.size - Nsites == 0:
            return None
        else:
            return Qubit(N=self.size - Nsites)

    def __repr__(self):
        return f"Qubit(N={self.size})"

    @property
    def _attrs(self):
        return (self.size,)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax

from netket.utils import timing
from netket.utils.api_utils import partial_from_kwargs
from netket import jax as nkjax
from netket.nn import split_array_mpi

from .qgt_jacobian_dense import QGTJacobianDenseT
from .qgt_jacobian_pytree import QGTJacobianPyTreeT
from .qgt_jacobian_common import (
    to_shift_offset,
    rescale,
)


@timing.timed
def QGTJacobian_DefaultConstructor(
    apply_fun,
    parameters,
    model_state,
    samples,
    pdf=None,
    *,
    dense: bool,
    mode: str | None = None,
    holomorphic: bool | None = None,
    diag_shift: float | None = 0.0,
    diag_scale: float | None = None,
    chunk_size: int | None = None,
    **kwargs,
) -> QGTJacobianDenseT | QGTJacobianPyTreeT:
    """
    Construct a :class:`QGTJacobianDenseT` or :class:`QGTJacobianPyTreeT`
    starting from the definition of a variational state.

    This type of the output is determined by the `dense` boolean keyword argument,
    which is mandatory.

    The `pdf` argument can be used to:
    - perform a full-summation calculation, where pdf is a probability density
      that sums to 1.
    - Similarly, it can be `jnp.ones_like(samples.shape[:-1])/Ns` to return
      the same result as `pdf = None`.
    - reweight individual samples a 'la importance sampling, where the pdf
      should now be the importance reweight factor times `1/Ns`.

    Other arguments are as other jacobian QGT constructors, documented below.

    This constructor is separated as it can be used from packages extending
    NetKet.
    """
    if mode is not None and holomorphic is not None:
        raise ValueError("Cannot specify both `mode` and `holomorphic`.")

    if mode is None:
        mode = nkjax.jacobian_default_mode(
            apply_fun,
            parameters,
            model_state,
            samples,
            holomorphic=holomorphic,
        )

    if pdf is not None:
        if not pdf.shape == samples.shape[:-1]:
            raise ValueError(
                "The shape of pdf must match the shape of the samples, "
                f"instead you provided (pdf.shape={pdf.shape}) != "
                f"(samples.shape={samples.shape[:-1]})"
            )
        if pdf.ndim >= 2:
            pdf = jax.jit(jax.lax.collapse, static_argnums=(1, 2))(pdf, 0, 2)

    if samples.ndim >= 3:
        # use jit so that we can do it on global shared array
        samples = jax.jit(jax.lax.collapse, static_argnums=(1, 2))(samples, 0, 2)

    jac_mode = mode
    if mode == "imag":
        # Imaginary mode is a specificity of the QGT, but it requires the standard complex-mode
        # jacobian to be computed.
        jac_mode = "complex"

    jacobians = nkjax.jacobian(
        apply_fun,
        parameters,
        samples,
        model_state,
        mode=jac_mode,
        pdf=pdf,
        chunk_size=chunk_size,
        dense=dense,
        center=True,
        _sqrt_rescale=True,
    )
    shift, offset = to_shift_offset(diag_shift, diag_scale)

    if offset is not None:
        ndims = 1 if (mode != "complex" and mode != "imag") else 2
        jacobians, scale = rescale(jacobians, offset, ndims=ndims)
    else:
        scale = None

    pars_struct = jax.tree_util.tree_map(
        lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), parameters
    )

    QGT_T = QGTJacobianDenseT if dense else QGTJacobianPyTreeT
    return QGT_T(
        O=jacobians,
        scale=scale,
        mode=mode,
        _params_structure=pars_struct,
        diag_shift=shift,
        **kwargs,
    )


@partial_from_kwargs(exclusive_arg_names=(("mode", "holomorphic")))
def QGTJacobianDense(
    vstate,
    *,
    mode: str | None = None,
    holomorphic: bool | None = None,
    diag_shift: float | None = 0.0,
    diag_scale: float | None = None,
    chunk_size: int | None = None,
    **kwargs,
) -> QGTJacobianDenseT:
    """
    Semi-lazy representation of an S Matrix where the Jacobian O_k is precomputed
    and stored as a dense matrix.

    The matrix of gradients O is computed on initialisation, but not S,
    which can be computed by calling :code:`to_dense`.
    The details on how the ⟨S⟩⁻¹⟨F⟩ system is solved are contained in
    the field `sr`.

    Numerical estimates of the QGT are usually ill-conditioned and require
    regularisation. The standard approach is to add a positive constant to the diagonal;
    alternatively, Becca and Sorella (2017) propose scaling this offset with the
    diagonal entry itself. NetKet allows using both in tandem:

    .. math::

        S_{ii} \\mapsto S_{ii} + \\epsilon_1 S_{ii} + \\epsilon_2;

    :math:`\\epsilon_{1,2}` are specified using `diag_scale` and `diag_shift`,
    respectively.

    Args:
        vstate: The variational state
        mode: "real", "complex" or "holomorphic": specifies the implementation
              used to compute the jacobian. "real" discards the imaginary part
              of the output of the model. "complex" splits the real and imaginary
              part of the parameters and output. It works also for non holomorphic
              models. holomorphic works for any function assuming it's holomorphic
              or real valued.
        holomorphic: a flag to indicate that the function is holomorphic.
        diag_scale: Fractional shift :math:`\\epsilon_1` added to diagonal entries (see above).
        diag_shift: Constant shift :math:`\\epsilon_2` added to diagonal entries (see above).
        chunk_size: If supplied, overrides the chunk size of the variational state
                    (useful for models where the backward pass requires more
                    memory than the forward pass).
    """
    # TODO: Find a better way to handle this case
    from netket.vqs import FullSumState

    if isinstance(vstate, FullSumState):
        samples = split_array_mpi(vstate._all_states)
        pdf = split_array_mpi(vstate.probability_distribution())
    else:
        samples = vstate.samples
        pdf = None

    if chunk_size is None:
        chunk_size = getattr(vstate, "chunk_size", None)

    return QGTJacobian_DefaultConstructor(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        samples,
        pdf=pdf,
        dense=True,
        mode=mode,
        holomorphic=holomorphic,
        diag_shift=diag_shift,
        diag_scale=diag_scale,
        chunk_size=chunk_size,
        **kwargs,
    )


@partial_from_kwargs(exclusive_arg_names=(("mode", "holomorphic")))
def QGTJacobianPyTree(
    vstate,
    *,
    mode: str | None = None,
    holomorphic: bool | None = None,
    diag_shift: float | None = 0.0,
    diag_scale: float | None = None,
    chunk_size: int | None = None,
    **kwargs,
) -> QGTJacobianPyTreeT:
    """
    Semi-lazy representation of an S Matrix where the Jacobian O_k is precomputed
    and stored as a PyTree.

    The matrix of gradients O is computed on initialisation, but not S,
    which can be computed by calling :code:`to_dense`.
    The details on how the ⟨S⟩⁻¹⟨F⟩ system is solved are contained in
    the field `sr`.

    Numerical estimates of the QGT are usually ill-conditioned and require
    regularisation. The standard approach is to add a positive constant to the diagonal;
    alternatively, Becca and Sorella (2017) propose scaling this offset with the
    diagonal entry itself. NetKet allows using both in tandem:

    .. math::

        S_{ii} \\mapsto S_{ii} + \\epsilon_1 S_{ii} + \\epsilon_2;

    :math:`\\epsilon_{1,2}` are specified using `diag_scale` and `diag_shift`,
    respectively.

    Args:
        vstate: The variational state
        mode: "real", "complex" or "holomorphic": specifies the implementation
              used to compute the jacobian. "real" discards the imaginary part
              of the output of the model. "complex" splits the real and imaginary
              part of the parameters and output. It works also for non holomorphic
              models. holomorphic works for any function assuming it's holomorphic
              or real valued.
        holomorphic: a flag to indicate that the function is holomorphic.
        diag_scale: Fractional shift :math:`\\epsilon_1` added to diagonal entries (see above).
        diag_shift: Constant shift :math:`\\epsilon_2` added to diagonal entries (see above).
        chunk_size: If supplied, overrides the chunk size of the variational state
                    (useful for models where the backward pass requires more
                    memory than the forward pass).
    """
    # TODO: Find a better way to handle this case
    from netket.vqs import FullSumState

    if isinstance(vstate, FullSumState):
        samples = split_array_mpi(vstate._all_states)
        pdf = split_array_mpi(vstate.probability_distribution())
    else:
        samples = vstate.samples
        pdf = None

    if chunk_size is None:
        chunk_size = getattr(vstate, "chunk_size", None)

    return QGTJacobian_DefaultConstructor(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        samples,
        dense=False,
        pdf=pdf,
        mode=mode,
        holomorphic=holomorphic,
        diag_shift=diag_shift,
        diag_scale=diag_scale,
        chunk_size=chunk_size,
        **kwargs,
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable
from collections.abc import Hashable

from netket.utils.types import DType, PyTree, Array

from netket.hilbert import AbstractHilbert
from netket.operator import ContinuousOperator
from netket.utils import struct, HashableArray

import jax
import jax.numpy as jnp


@struct.dataclass
class PotentialOperatorPyTree:
    """Internal class used to pass data from the operator to the jax kernel.

    This is used such that we can pass a PyTree containing some static data.
    We could avoid this if the operator itself was a pytree, but as this is not
    the case we need to pass as a separte object all fields that are used in
    the kernel.

    We could forego this, but then the kernel could not be marked as
    @staticmethod and we would recompile every time we construct a new operator,
    even if it is identical
    """

    potential_fun: Callable = struct.field(pytree_node=False)
    coefficient: Array


class PotentialEnergy(ContinuousOperator):
    r"""Returns the local potential energy defined in afun"""

    def __init__(
        self,
        hilbert: AbstractHilbert,
        afun: Callable,
        coefficient: float = 1.0,
        dtype: DType | None = None,
    ):
        r"""
        Args:
            hilbert: The underlying Hilbert space on which the operator is defined
            afun: The potential energy as function of x
            coefficient: A coefficient for the ContinuousOperator object
            dtype: Data type of the coefficient
        """

        self._afun = afun
        self._coefficient = jnp.array(coefficient, dtype=dtype)

        self.__attrs = None

        super().__init__(hilbert, self._coefficient.dtype)

    @property
    def coefficient(self) -> Array:
        return self._coefficient

    @property
    def is_hermitian(self) -> bool:
        return True

    @staticmethod
    def _expect_kernel(
        logpsi: Callable, params: PyTree, x: Array, data: PyTree | None
    ) -> Array:
        return data.coefficient * jax.vmap(data.potential_fun, in_axes=(0,))(x)

    def _pack_arguments(self) -> PotentialOperatorPyTree:
        return PotentialOperatorPyTree(self._afun, self.coefficient)

    @property
    def _attrs(self) -> tuple[Hashable, ...]:
        if self.__attrs is None:
            self.__attrs = (
                self.hilbert,
                self._afun,
                self.dtype,
                HashableArray(self.coefficient),
            )
        return self.__attrs

    def __repr__(self):
        return f"Potential(coefficient={self.coefficient}, function={self._afun})"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce

import numpy as np
from jax import numpy as jnp

from netket.graph import AbstractGraph
from netket.hilbert import Fock
from netket.jax import canonicalize_dtypes
from netket.utils.numbers import dtype as _dtype
from netket.utils.types import DType

from .. import boson
from .._hamiltonian import SpecialHamiltonian
from .._local_operator import LocalOperator


class BoseHubbardBase(SpecialHamiltonian):
    r"""
    An extended Bose Hubbard model Hamiltonian operator, containing both
    on-site interactions and nearest-neighboring density-density interactions.
    """

    def __init__(
        self,
        hilbert: Fock,
        graph: AbstractGraph,
        U: float,
        V: float = 0.0,
        J: float = 1.0,
        mu: float = 0.0,
        dtype: DType | None = None,
    ):
        r"""
        Constructs a new BoseHubbard operator given a hilbert space, a graph
        specifying the connectivity and the interaction strength.
        The chemical potential and the density-density interaction strength
        can be specified as well.

        Args:
           hilbert: Hilbert space the operator acts on.
           U: The on-site interaction term.
           V: The strength of density-density interaction term.
           J: The hopping amplitude.
           mu: The chemical potential.
           dtype: The dtype of the matrix elements.

        Examples:
           Constructs a BoseHubbard operator for a 2D system.

           >>> import netket as nk
           >>> g = nk.graph.Hypercube(length=3, n_dim=2, pbc=True)
           >>> hi = nk.hilbert.Fock(n_max=3, n_particles=6, N=g.n_nodes)
           >>> op = nk.operator.BoseHubbard(hi, U=4.0, graph=g)
           >>> print(op.hilbert.size)
           9
        """
        assert isinstance(hilbert, Fock)
        super().__init__(hilbert)

        if isinstance(graph, AbstractGraph):
            if graph.n_nodes != hilbert.size:
                raise ValueError(
                    """
                    The size of the graph must match the hilbert space.
                    """
                )
            # support also a matrix input in here.
            graph = graph.edges()

        if isinstance(graph, list):
            graph = np.asarray(
                [[u, v] for u, v in graph],
                dtype=np.intp,
            )

        if graph.ndim != 2 or graph.shape[1] != 2:
            raise ValueError(
                """
                Graph should be one of:
                    - NetKet graph type (nk.operator.AbstractGraph)
                    - List of tuples, describing the edges
                    - a (N,2) array of integers.
                """
            )

        dtype = canonicalize_dtypes(float, U, V, J, mu, dtype=dtype)
        self._U = U.astype(dtype=dtype)
        self._V = V.astype(dtype=dtype)
        self._J = J.astype(dtype=dtype)
        self._mu = mu.astype(dtype=dtype)

        self._n_max = hilbert.n_max
        self._n_sites = hilbert.size
        self._edges = graph.astype(np.intp)
        self._max_conn = 1 + self._edges.shape[0] * 2

    @property
    def is_hermitian(self):
        return True

    @property
    def dtype(self) -> DType:
        """The dtype of the matrix elements."""
        return reduce(
            jnp.promote_types,
            (_dtype(self.U), _dtype(self.V), _dtype(self._J), _dtype(self._mu)),
        )

    @property
    def edges(self) -> np.ndarray:
        return self._edges

    @property
    def U(self):
        """The strength of on-site interaction term."""
        return self._U

    @property
    def V(self):
        """The strength of density-density interaction term."""
        return self._V

    @property
    def J(self):
        """The hopping amplitude."""
        return self._J

    @property
    def mu(self):
        """The chemical potential."""
        return self._mu

    def conjugate(self, *, concrete=True):
        # if real
        if isinstance(self.dtype, float):
            return self
        else:
            raise NotImplementedError

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        return self._max_conn

    def copy(self, *, dtype: DType | None = None):
        if dtype is None:
            dtype = self.dtype

        return type(self)(
            hilbert=self.hilbert,
            graph=self.edges,
            U=self.U,
            V=self.V,
            J=self.J,
            mu=self.mu,
            dtype=dtype,
        )

    def to_local_operator(self):
        # The hamiltonian
        ha = LocalOperator(self.hilbert, dtype=self.dtype)

        if self.U != 0 or self.mu != 0:
            for i in range(self.hilbert.size):
                n_i = boson.number(self.hilbert, i)
                ha += (self.U / 2) * n_i * (n_i - 1) - self.mu * n_i

        if self.J != 0:
            for i, j in self.edges:
                ha += self.V * (
                    boson.number(self.hilbert, int(i))
                    * boson.number(self.hilbert, int(j))
                )
                ha -= self.J * (
                    boson.destroy(self.hilbert, int(i))
                    * boson.create(self.hilbert, int(j))
                    + boson.create(self.hilbert, int(i))
                    * boson.destroy(self.hilbert, int(j))
                )

        return ha

    def _iadd_same_hamiltonian(self, other):
        if self.hilbert != other.hilbert:
            raise NotImplementedError(
                "Cannot add hamiltonians on different hilbert spaces"
            )

        self._U += other.U
        self._V += other.V
        self._J += other.J
        self._mu += other.mu

    def _isub_same_hamiltonian(self, other):
        if self.hilbert != other.hilbert:
            raise NotImplementedError(
                "Cannot add hamiltonians on different hilbert spaces"
            )

        self._U -= other.U
        self._V -= other.V
        self._J -= other.J
        self._mu -= other.mu

    def __repr__(self):
        return f"{type(self).__name__}(U={self._U}, V={self._V}, J={self._J}, mu={self._mu}; dim={self.hilbert.size})"



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import rules


# Deprecated bindings from stabilisaation
from netket.sampler import (
    ParallelTemperingSampler as _deprecated_MetropolisPtSampler,
    ParallelTemperingLocal as _deprecated_MetropolisLocalPtSampler,
    ParallelTemperingExchange as _deprecated_MetropolisExchangePtSampler,
    MetropolisFermionHop as _deprecated_MetropolisParticleExchange,
    MetropolisSampler as _deprecated_MetropolisSampler,
)

_deprecations = {
    # May 2024, NetKet 3.12
    "MetropolisPtSampler": (
        "netket.experimental.sampler.MetropolisPtSampler is deprecated: use "
        "netket.sampler.ParallelTemperingSampler (netket >= 3.12)",
        _deprecated_MetropolisPtSampler,
    ),
    "MetropolisLocalPt": (
        "netket.experimental.sampler.MetropolisLocalPt is deprecated: use "
        "netket.sampler.ParallelTemperingLocal (netket >= 3.12)",
        _deprecated_MetropolisLocalPtSampler,
    ),
    "MetropolisExchangePt": (
        "netket.experimental.sampler.MetropolisExchangePt is deprecated: use "
        "netket.sampler.ParallelTemperingExchange (netket >= 3.12)",
        _deprecated_MetropolisExchangePtSampler,
    ),
    # June 2024, NetKet 3.13
    "MetropolisParticleExchange": (
        "netket.experimental.sampler.MetropolisParticleExchange is deprecated: use "
        "netket.sampler.MetropolisFermionHop (netket >= 3.13)",
        _deprecated_MetropolisParticleExchange,
    ),
    # June 2025
    "MetropolisSamplerPmap": (
        "netket.experimental.sampler.MetropolisSamplerPmap is deprecated: just use "
        "netket.sampler.MetropolisSampler",
        _deprecated_MetropolisSampler,
    ),
}


from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr
from netket.utils import _hide_submodules

__getattr__ = _deprecation_getattr(__name__, _deprecations)
_hide_submodules(__name__)

del _deprecation_getattr



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
import abc

from flax import linen as nn
from jax import numpy as jnp

from netket.utils.types import PyTree, PRNGKeyT

from netket.utils import struct
from netket import config

# Necessary for the type annotation to work
if config.netket_sphinx_build:
    from netket import sampler


class MetropolisRule(struct.Pytree):
    """
    Base class for transition rules of Metropolis, such as Local, Exchange, Hamiltonian
    and several others.
    """

    def init_state(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        key: PRNGKeyT,
    ) -> Any | None:
        """
        Initialises the optional internal state of the Metropolis sampler transition
        rule.

        The provided key is unique and does not need to be split.

        It should return an immutable data structure.

        Arguments:
            sampler: The Metropolis sampler.
            machine: A Flax module with the forward pass of the log-pdf.
            params: The PyTree of parameters of the model.
            key: A Jax PRNGKey.

        Returns:
            An optional state.
        """
        return None

    def reset(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        sampler_state: "sampler.SamplerState",  # noqa: F821
    ) -> Any | None:
        """
        Resets the internal state of the Metropolis Sampler Transition Rule.

        The default implementation returns the current rule_state without modifying it.

        Arguments:
            sampler: The Metropolis sampler.
            machine: A Flax module with the forward pass of the log-pdf.
            params: The PyTree of parameters of the model.
            sampler_state: The current state of the sampler. Should not modify it.

        Returns:
           A reset state of the rule. This returns the same type of
           :py:meth:`~nk.sampler.rule.MetropolisRule.rule_state` and might be `None`.
        """
        return sampler_state.rule_state

    @abc.abstractmethod
    def transition(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        sampler_state: "sampler.SamplerState",  # noqa: F821
        key: PRNGKeyT,
        σ: jnp.ndarray,
    ) -> tuple[jnp.ndarray, jnp.ndarray | None]:
        r"""
        Proposes a new configuration set of configurations $\sigma'$ starting from the current
        chain configurations :math:`\sigma`.

        The new configurations :math:`\sigma'` should be a matrix with the same dimension as
        :math:`\sigma`.

        This function should return a tuple. where the first element are the new configurations
        $\sigma'$ and the second element is either `None` or an array of length `σ.shape[0]`
        containing an optional log-correction factor. The correction factor should be non-zero
        when the transition rule is non-symmetrical.

        Arguments:
            sampler: The Metropolis sampler.
            machine: A Flax module with the forward pass of the log-pdf.
            params: The PyTree of parameters of the model.
            sampler_state: The current state of the sampler. Should not modify it.
            key: A Jax PRNGKey to use to generate new random configurations.
            σ: The current configurations stored in a 2D matrix.

        Returns:
           A tuple containing the new configurations :math:`\sigma'` and the optional vector of
           log corrections to the transition probability.
        """

    def random_state(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        sampler_state: "sampler.SamplerState",  # noqa: F821
        key: PRNGKeyT,
    ):
        """
        Generates a random state compatible with this rule.

        By default this calls :func:`netket.hilbert.random.random_state`.

        Arguments:
            sampler: The Metropolis sampler.
            machine: A Flax module with the forward pass of the log-pdf.
            params: The PyTree of parameters of the model.
            sampler_state: The current state of the sampler. Should not modify it.
            key: The PRNGKey to use to generate the random state.
        """
        return sampler.hilbert.random_state(
            key, size=sampler.n_batches, dtype=sampler.dtype
        )



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flax import linen as nn
from flax.linen.dtypes import promote_dtype
from jax import numpy as jnp
from jax.nn.initializers import zeros

from netket.utils.types import Array

from .layers import RNNLayer


class FastRNNLayer(RNNLayer):
    """
    Recurrent neural network layer with fast sampling.

    See :class:`netket.models.FastARNNSequential` for a brief explanation of fast autoregressive sampling.
    """

    size: int = None
    """number of sites."""

    @nn.compact
    def update_site(self, inputs: Array, index: int) -> Array:
        """
        Applies the RNN cell to a batch of input sites at a given index,
        and stores the updated memories in the cache.

        Args:
          inputs: an input site with dimensions (batch, features).
          index: the index of the output site. The index of the input site should be `index - self.exclusive`.

        Returns:
          The output site with dimensions (batch, features).
        """
        batch_size = inputs.shape[0]
        inputs = promote_dtype(inputs, dtype=self.cell.param_dtype)[0]

        if self.reorder_idx is None:
            prev_neighbors = None
        else:
            prev_neighbors = jnp.asarray(self.prev_neighbors)

        _cell_mem = self.variable(
            "cache",
            "cell_mem",
            zeros,
            None,
            (batch_size, self.cell.features),
            inputs.dtype,
        )
        _outputs = self.variable(
            "cache",
            "outputs",
            zeros,
            None,
            (batch_size, self.size, self.cell.features),
            inputs.dtype,
        )
        cell_mem = _cell_mem.value
        outputs = _outputs.value

        hidden = self._extract_hidden(outputs, index, prev_neighbors)
        cell_mem, hidden = self.cell(inputs, cell_mem, hidden)

        if not self.is_initializing():
            _cell_mem.value = cell_mem
            _outputs.value = outputs.at[:, index, :].set(hidden)

        return hidden

    def __call__(self, inputs: Array) -> Array:
        return RNNLayer.__call__(self, inputs)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scipy import sparse as _sparse

from netket.utils.types import DType as _DType

from netket.hilbert import DiscreteHilbert as _DiscreteHilbert

from ._local_operator import LocalOperator as _LocalOperator


def identity(hilbert: _DiscreteHilbert, dtype: _DType = None) -> _LocalOperator:
    """
    Builds the :math:`\\mathbb{I}` identity operator.

    Args:
        hilbert: The hilbert space.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    return _LocalOperator(hilbert, constant=1.0, dtype=dtype)


def sigmax(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = None
) -> _LocalOperator:
    """
    Builds the :math:`\\sigma^x` operator acting on the `site`-th of the Hilbert
    space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)
    S = (N - 1) / 2

    D = [np.sqrt((S + 1) * 2 * a - a * (a + 1)) for a in np.arange(1, N)]
    mat = np.diag(D, 1) + np.diag(D, -1)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)


def sigmay(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = None
) -> _LocalOperator:
    """
    Builds the :math:`\\sigma^y` operator acting on the `site`-th of the Hilbert
    space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np
    import netket.jax as nkjax
    from netket.utils import module_version

    if module_version(np) >= (2, 0, 0):
        from numpy.exceptions import ComplexWarning
    else:
        from numpy import ComplexWarning  # type: ignore

    if not nkjax.is_complex_dtype(dtype):
        import jax.numpy as jnp
        import warnings

        old_dtype = dtype
        dtype = jnp.promote_types(complex, old_dtype)
        if old_dtype is not None:
            warnings.warn(
                ComplexWarning(
                    f"A complex dtype is required (dtype={old_dtype} specified). "
                    f"Promoting to dtype={dtype}."
                ),
                stacklevel=2,
            )

    N = hilbert.size_at_index(site)
    S = (N - 1) / 2

    D = np.array([1j * np.sqrt((S + 1) * 2 * a - a * (a + 1)) for a in np.arange(1, N)])
    mat = np.diag(D, -1) + np.diag(-D, 1)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)


def sigmaz(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = None
) -> _LocalOperator:
    """
    Builds the :math:`\\sigma^z` operator acting on the `site`-th of the Hilbert
    space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)
    S = (N - 1) / 2

    D = np.array([2 * m for m in np.arange(S, -(S + 1), -1)])
    mat = np.diag(D, 0)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)


def sigmam(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = None
) -> _LocalOperator:
    """
    Builds the :math:`\\sigma^{-} = \\frac{1}{2}(\\sigma^x - i \\sigma^y)` operator acting on the
    `site`-th of the Hilbert space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)
    S = (N - 1) / 2

    S2 = (S + 1) * S
    D = np.array([np.sqrt(S2 - m * (m - 1)) for m in np.arange(S, -S, -1)])
    mat = np.diag(D, -1)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)


def sigmap(
    hilbert: _DiscreteHilbert, site: int, dtype: _DType = None
) -> _LocalOperator:
    """
    Builds the :math:`\\sigma^{+} = \\frac{1}{2}(\\sigma^x + i \\sigma^y)` operator acting on the
    `site`-th of the Hilbert space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    Args:
        hilbert: The hilbert space.
        site: The site on which this operator acts.
        dtype: The datatype to use for the matrix elements.

    Returns:
        An instance of {class}`nk.operator.LocalOperator`.
    """
    import numpy as np

    N = hilbert.size_at_index(site)
    S = (N - 1) / 2

    S2 = (S + 1) * S
    D = np.array([np.sqrt(S2 - m * (m + 1)) for m in np.arange(S - 1, -(S + 1), -1)])
    mat = np.diag(D, 1)
    mat = _sparse.coo_matrix(mat)
    return _LocalOperator(hilbert, mat, [site], dtype=dtype)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from dataclasses import dataclass
from functools import partial, wraps

from typing import Any

import numpy as np
from numba import jit
from jax import numpy as jnp
import jax

from netket.hilbert import AbstractHilbert
from netket.utils.mpi import mpi_sum, n_nodes
from netket.utils.types import PyTree, ModuleOrApplyFun
from netket import config

import netket.jax as nkjax

from .metropolis import MetropolisSampler, MetropolisRule


@dataclass
class MetropolisNumpySamplerState:
    σ: np.ndarray
    """Holds the current configuration."""
    σ1: np.ndarray
    """Holds a proposed configuration (preallocation)."""

    log_prob: np.ndarray
    """Holds model(pars, σ) for the current σ (preallocation)."""
    log_prob_corr: np.ndarray
    """Holds optional acceptance correction (preallocation)."""

    rule_state: Any
    """The optional state of the rule."""
    rng: Any
    """A numpy random generator."""

    n_steps_proc: int = 0
    """Number of moves performed along the chains in this process since the last reset."""
    n_accepted_proc: int = 0
    """Number of accepted transitions among the chains in this process since the last reset."""

    @property
    def acceptance(self) -> float | None:
        """The fraction of accepted moves across all chains and MPI processes.

        The rate is computed since the last reset of the sampler.
        Will return None if no sampling has been performed since then.
        """
        if self.n_steps == 0:
            return None

        return self.n_accepted / self.n_steps

    @property
    def n_steps(self) -> int:
        """Total number of moves performed across all processes since the last reset."""
        return self.n_steps_proc * n_nodes

    @property
    def n_accepted(self) -> int:
        """Total number of moves accepted across all processes since the last reset."""
        return mpi_sum(self.n_accepted_proc)

    def __repr__(self):
        if self.n_steps > 0:
            acc_string = f"# accepted = {self.n_accepted}/{self.n_steps} ({self.acceptance * 100}%), "
        else:
            acc_string = ""

        return f"MetropolisNumpySamplerState({acc_string}rng state={self.rng})"


@partial(jax.jit, static_argnums=(0, 3))
def apply_model(machine, pars, weights, chunk_size):
    chunked = nkjax.apply_chunked(
        machine.apply, in_axes=(None, 0), chunk_size=chunk_size
    )
    return chunked(pars, weights)


class MetropolisSamplerNumpy(MetropolisSampler):
    """
    Metropolis-Hastings sampler for an Hilbert space according to a specific transition
    rule executed on CPU through Numpy.

    This sampler is equivalent to :ref:`netket.sampler.MetropolisSampler` but instead of
    executing the whole sampling inside a jax-jitted function, only evaluates the forward
    pass inside a jax-jitted function, while proposing new steps and accepting/rejecting
    them is performed in numpy.

    Because of Jax dispatch cost, and especially for small system, this sampler performs
    poorly, while asymptotically it should have the same performance of standard Jax samplers.

    However, some transition rules don't work on GPU, and some samplers (Hamiltonian) work
    very poorly on jax so this is a good workaround.

    See :ref:`netket.sampler.MetropolisSampler` for more information.

    .. warning::
       This sampler only works on the CPU. To use the Hamiltonian sampler with GPUs, you
       should use :class:`netket.sampler.MetropolisHamiltonian`.

       This sampler is largely outdated, and we recommend not to use them. Little effort has
       been put into improving performance in those samplers after 2021.

       If you have a complicated transitions rule, consider instead using a `jax.pure_callback`.
    """

    @wraps(MetropolisSampler.__init__)
    def __init__(self, hilbert: AbstractHilbert, rule: MetropolisRule, **kwargs):
        super().__init__(hilbert, rule, **kwargs)
        # standard samplers use jax arrays, this must be a numpy array
        self.machine_pow = np.array(self.machine_pow)

    @property
    def n_batches(self) -> int:
        r"""
        The batch size of the configuration $\sigma$ used by this sampler on this
        jax process.

        This is used to determine the shape of the batches generated in a single process.
        This is needed because when using MPI, every process must create a batch of chains
        of :attr:`~Sampler.n_chains_per_rank`, while when using the experimental sharding
        mode we must declare the full shape on every jax process, therefore this returns
        :attr:`~Sampler.n_chains`.

        Usage of this flag is required to support both MPI and sharding.

        Samplers may override this to have a larger batch size, for example to
        propagate multiple replicas (in the case of parallel tempering).
        """
        return self.n_chains_per_rank

    def _init_state(self, machine, parameters, key):
        rgen = np.random.default_rng(np.asarray(key))

        σ = np.zeros((self.n_batches, self.hilbert.size), dtype=self.dtype)

        ma_out = jax.eval_shape(machine.apply, parameters, σ)

        state = MetropolisNumpySamplerState(
            σ=σ,
            σ1=np.copy(σ),
            log_prob=np.zeros(self.n_batches, dtype=ma_out.dtype),
            log_prob_corr=np.zeros(
                self.n_batches, dtype=nkjax.dtype_real(ma_out.dtype)
            ),
            rng=rgen,
            rule_state=self.rule.init_state(self, machine, parameters, rgen),
        )

        if not self.reset_chains:
            key = jnp.asarray(
                state.rng.integers(0, 1 << 32, size=2, dtype=np.uint32), dtype=np.uint32
            )

            state.σ = np.copy(
                self.rule.random_state(self, machine, parameters, state, key)
            )

        return state

    def _reset(self, machine, parameters, state):
        if self.reset_chains:
            # directly generate a PRNGKey which is a [2xuint32] array
            key = jnp.asarray(
                state.rng.integers(0, 1 << 32, size=2, dtype=np.uint32), dtype=np.uint32
            )
            state.σ = np.copy(
                self.rule.random_state(self, machine, parameters, state, key)
            )

        state.rule_state = self.rule.reset(self, machine, parameters, state)
        state.log_prob = np.array(
            self.machine_pow
            * apply_model(machine, parameters, state.σ, self.chunk_size).real
        )

        state._accepted_samples = 0
        state._total_samples = 0

        return state

    def _sample_next(
        self,
        machine: ModuleOrApplyFun,
        parameters: PyTree,
        state: MetropolisNumpySamplerState,
    ) -> tuple[MetropolisNumpySamplerState, tuple[np.ndarray, np.ndarray]]:
        σ = state.σ
        σ1 = state.σ1
        log_prob = state.log_prob
        log_prob_corr = state.log_prob_corr
        mpow = self.machine_pow

        rgen = state.rng

        accepted = 0

        for sweep in range(self.sweep_size):
            # Propose a new state using the transition kernel
            # σp, log_prob_correction =
            self.rule.transition(self, machine, parameters, state, state.rng, σ)

            if config.netket_experimental_sharding:
                from jax.experimental.multihost_utils import (
                    host_local_array_to_global_array,
                    global_array_to_host_local_array,
                )

                global_mesh = jax.sharding.Mesh(jax.devices(), "x")
                pspecs = jax.sharding.PartitionSpec("x")

                all_samples = host_local_array_to_global_array(σ1, global_mesh, pspecs)
                _log_prob = (
                    mpow
                    * apply_model(
                        machine, parameters, all_samples, self.chunk_size
                    ).real
                )
                assert len(_log_prob.addressable_shards) == 1
                _log_prob = global_array_to_host_local_array(
                    log_prob, global_mesh, pspecs
                )
            else:
                _log_prob = (
                    mpow * apply_model(machine, parameters, σ1, self.chunk_size).real
                )

            log_prob_1 = np.copy(_log_prob)
            assert log_prob_1.shape == σ1.shape[:-1]

            random_uniform = rgen.uniform(0, 1, size=σ.shape[0])

            # Acceptance Kernel
            accepted += acceptance_kernel(
                σ,
                σ1,
                log_prob,
                log_prob_1,
                log_prob_corr,
                mpow,
                random_uniform,
            )

        state.n_steps_proc += self.sweep_size * self.n_chains_per_rank
        state.n_accepted_proc += accepted

        return state, (state.σ, state.log_prob)

    def _sample_chain(
        self,
        machine: ModuleOrApplyFun,
        parameters: PyTree,
        state: MetropolisNumpySamplerState,
        chain_length: int,
        return_log_probabilities: bool = False,
    ) -> tuple[jnp.ndarray, MetropolisNumpySamplerState]:
        samples = np.empty(
            (chain_length, self.n_chains_per_rank, self.hilbert.size),
            dtype=self.dtype,
        )
        log_probs = np.empty((chain_length, self.n_chains_per_rank), dtype=float)

        for i in range(chain_length):
            state, (σ, log_prob) = self.sample_next(machine, parameters, state)
            samples[i] = σ
            log_probs[i] = log_prob

        # make it (n_chains_per_rank, n_samples_per_chain, hi.size) as expected by netket.stats.statistics
        samples = np.swapaxes(samples, 0, 1)
        log_probs = np.swapaxes(log_probs, 0, 1)

        if return_log_probabilities:
            return (samples, log_probs), state
        else:
            return samples, state

    def __repr__(self):
        return (
            "MetropolisSamplerNumpy("
            + f"\n  hilbert = {self.hilbert},"
            + f"\n  rule = {self.rule},"
            + f"\n  n_chains = {self.n_chains},"
            + f"\n  machine_power = {self.machine_pow},"
            + f"\n  reset_chains = {self.reset_chains},"
            + f"\n  sweep_size = {self.sweep_size},"
            + f"\n  dtype = {self.dtype},"
            + ")"
        )

    def __str__(self):
        return (
            "MetropolisSamplerNumpy("
            + f"rule = {self.rule}, "
            + f"n_chains = {self.n_chains}, "
            + f"machine_power = {self.machine_pow}, "
            + f"sweep_size = {self.sweep_size}, "
            + f"dtype = {self.dtype})"
        )


@jit(nopython=True)
def acceptance_kernel(
    σ, σ1, log_prob, log_prob_1, log_prob_corr, machine_pow, random_uniform
):
    accepted = 0

    for i in range(σ.shape[0]):
        prob = np.exp(log_prob_1[i] - log_prob[i] + log_prob_corr[i])
        assert not math.isnan(prob)

        if prob > random_uniform[i]:
            log_prob[i] = log_prob_1[i]
            σ[i] = σ1[i]
            accepted += 1

    return accepted


def MetropolisLocalNumpy(hilbert: AbstractHilbert, **kwargs):
    from .rules import LocalRuleNumpy

    rule = LocalRuleNumpy()
    return MetropolisSamplerNumpy(hilbert, rule, **kwargs)


def MetropolisHamiltonianNumpy(hilbert: AbstractHilbert, hamiltonian, **kwargs):
    from .rules import HamiltonianRuleNumpy

    rule = HamiltonianRuleNumpy(hamiltonian)
    return MetropolisSamplerNumpy(hilbert, rule, **kwargs)


def MetropolisCustomNumpy(
    hilbert: AbstractHilbert, move_operators, move_weights=None, **kwargs
):
    from .rules import CustomRuleNumpy

    rule = CustomRuleNumpy(move_operators, move_weights)
    return MetropolisSamplerNumpy(hilbert, rule, **kwargs)



from collections.abc import Callable
from typing import TYPE_CHECKING

import jax
from jax import numpy as jnp
from flax import linen as nn
from netket.utils.types import NNInitFunc, DType
from jax.nn.initializers import (
    zeros,
    ones,
    lecun_normal,
)

import netket.nn as nknn

if TYPE_CHECKING:
    from netket.hilbert import ContinuousHilbert


class DeepSetMLP(nn.Module):
    r"""Implements the DeepSets architecture, which is permutation invariant.

    .. math ::

        f(x_1,...,x_N) = \rho\left(\sum_i \phi(x_i)\right)

    that is suitable for the simulation of bosonic.

    The input shape must have an axis that is reshaped to (..., N, D), where we pool over N.

    See DeepSetRelDistance for the bosonic wave function ansatz in
    https://arxiv.org/abs/1703.06114
    """

    features_phi: int | tuple[int, ...] | None = None
    """
    Number of features in each layer for phi network.
    When features_phi is None, no phi network is created.
    """
    features_rho: int | tuple[int, ...] | None = None
    """
    Number of features in each layer for rho network.
    Should not include the final layer of dimension 1, which is included automatically.
    When features_rho is None, a single layer MLP with output 1 is created.
    """

    param_dtype: DType = jnp.float64
    """The dtype of the weights."""

    hidden_activation: Callable | None = jax.nn.gelu
    """The nonlinear activation function between hidden layers."""
    output_activation: Callable | None = None
    """The nonlinear activation function at the output layer."""

    pooling: Callable = jnp.sum
    """The pooling operation to be used after the phi-transformation"""

    use_bias: bool = True
    """if True uses a bias in all layers."""

    kernel_init: NNInitFunc = lecun_normal()
    """Initializer for the Dense layer matrix"""
    bias_init: NNInitFunc = zeros
    """Initializer for the hidden bias"""
    precision: jax.lax.Precision | None = None
    """numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    @nn.compact
    def __call__(self, input):
        features_rho = _process_features_rho(self.features_rho)
        ds = nknn.blocks.DeepSetMLP(
            features_phi=self.features_phi,
            features_rho=features_rho,
            param_dtype=self.param_dtype,
            hidden_activation=self.hidden_activation,
            output_activation=self.output_activation,
            pooling=self.pooling,
            use_bias=self.use_bias,
            kernel_init=self.kernel_init,
            bias_init=self.bias_init,
            precision=self.precision,
        )
        x = ds(input)
        x = x.squeeze(-1)
        return x


def _process_features_rho(features_input):
    if features_input is None:
        return (1,)
    elif isinstance(features_input, int):
        return (features_input, 1)
    else:
        if not hasattr(features_input, "__len__"):
            raise ValueError("features_rho must be a sequence of integers")
        return (
            *tuple(features_input),
            1,
        )


class DeepSetRelDistance(nn.Module):
    r"""Implements an equivariant version of the DeepSets architecture
    given by (https://arxiv.org/abs/1703.06114)

    .. math ::

        f(x_1,...,x_N) = \rho\left(\sum_i \phi(x_i)\right)

    that is suitable for the simulation of periodic systems.
    Additionally one can add a cusp condition by specifying the
    asymptotic exponent.
    For helium the Ansatz reads (https://arxiv.org/abs/2112.11957):

    .. math ::

        \psi(x_1,...,x_N) = \rho\left(\sum_i \phi(d_{\sin}(x_i,x_j))\right) \cdot \exp\left[-\frac{1}{2}\left(b/d_{\sin}(x_i,x_j)\right)^5\right]

    """

    hilbert: "ContinuousHilbert"
    """The hilbert space defining the periodic box where this ansatz is defined."""

    layers_phi: int
    """Number of layers in phi network."""
    layers_rho: int
    """Number of layers in rho network."""

    features_phi: tuple | int
    """Number of features in each layer for phi network."""
    features_rho: tuple | int
    """
    Number of features in each layer for rho network.
    If specified as a list, the last layer must have 1 feature.
    """

    cusp_exponent: int | None = None
    """exponent of Katos cusp condition"""

    param_dtype: DType = jnp.float64
    """The dtype of the weights."""

    activation: Callable | None = jax.nn.gelu
    """The nonlinear activation function between hidden layers."""
    output_activation: Callable | None = None
    """The nonlinear activation function at the output layer."""

    pooling: Callable = jnp.sum
    """The pooling operation to be used after the phi-transformation"""

    use_bias: bool = True
    """if True uses a bias in all layers."""

    kernel_init: NNInitFunc = lecun_normal()
    """Initializer for the Dense layer matrix"""
    bias_init: NNInitFunc = zeros
    """Initializer for the hidden bias"""
    params_init: NNInitFunc = ones
    """Initializer for the parameter in the cusp"""

    def setup(self):
        if not all(self.hilbert.geometry.pbc):
            raise ValueError(
                "The DeepSetRelDistance model only works with "
                "hilbert spaces with periodic boundary conditions "
                "among all directions."
            )

        features_phi = self.features_phi
        if isinstance(features_phi, int):
            features_phi = [features_phi] * self.layers_phi
        features_phi = tuple(features_phi)

        check_features_length(features_phi, self.layers_phi, "phi")

        features_rho = self.features_rho
        if isinstance(features_rho, int):
            features_rho = [features_rho] * (self.layers_rho - 1) + [1]
        features_rho = tuple(features_rho)

        check_features_length(features_rho, self.layers_rho, "rho")
        assert features_rho[-1] == 1

        self.deepset = DeepSetMLP(
            features_phi,
            features_rho,
            param_dtype=self.param_dtype,
            hidden_activation=self.activation,
            output_activation=self.output_activation,
            pooling=self.pooling,
            use_bias=self.use_bias,
            kernel_init=self.kernel_init,
            bias_init=self.bias_init,
        )

    def distance(self, x, sdim, L):
        n_particles = x.shape[0] // sdim
        x = x.reshape(-1, sdim)

        dis = -x[jnp.newaxis, :, :] + x[:, jnp.newaxis, :]

        dis = dis[jnp.triu_indices(n_particles, 1)]
        dis = L[jnp.newaxis, :] / 2.0 * jnp.sin(jnp.pi * dis / L[jnp.newaxis, :])
        return dis

    @nn.compact
    def __call__(self, x):
        batch_shape = x.shape[:-1]
        param = self.param("cusp", self.params_init, (1,), self.param_dtype)

        L = jnp.array(self.hilbert.domain)
        sdim = L.size

        d = jax.vmap(self.distance, in_axes=(0, None, None))(x, sdim, L)
        dis = jnp.linalg.norm(d, axis=-1)

        cusp = 0.0
        if self.cusp_exponent is not None:
            cusp = -0.5 * jnp.sum(param / dis**self.cusp_exponent, axis=-1)

        y = (d / L[jnp.newaxis, :]) ** 2

        y = self.deepset(y)

        return (y + cusp).reshape(*batch_shape)


def check_features_length(features, n_layers, name):
    if len(features) != n_layers:
        raise ValueError(
            f"The number of {name} layers ({n_layers}) does not match "
            f"the length of the features list ({len(features)})."
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Literal, Union, TYPE_CHECKING

import tarfile
import time
from io import BytesIO
import glob

import os
from os import path as _path

from flax import serialization

from netket.jax.sharding import extract_replicated

if TYPE_CHECKING:
    from netket.vqs import VariationalState

from .base import AbstractLog

FileModeT = Union[
    Literal["write"],
    Literal["append"],
    Literal["fail"],
    Literal["x"],
    Literal["w"],
    Literal["a"],
]


def save_binary_to_tar(tar_file, byte_data, name):
    abuf = BytesIO(byte_data)

    # Construct the info object with the correct length
    info = tarfile.TarInfo(name=name)
    info.size = len(abuf.getbuffer())

    # actually save the data to the tar file
    tar_file.addfile(tarinfo=info, fileobj=abuf)


class StateLog(AbstractLog):
    """
    A logger which serializes the variables of the variational state during a run.

    The data is saved either to a directory or tar archive in a sequence of files named
    `[0.mpack, 1.mpack, ...]` where the filename is incremented every time the logger is
    called. The tar file inside is not flushed to disk (closed) until this object is
    deleted or python is shut down.
    """

    def __init__(
        self,
        output_prefix: str,
        mode: FileModeT = "write",
        save_every: int = 1,
        tar: bool = False,
    ):
        """
        Initialize the :code:`StateLogger`.

        Args:
            output_prefix: the name of the output file before the extension (if
                tar=True) or of the output folder.
            save_every: every how many iterations the variables should be saved.
                (default 1)
            mode: Specify the behaviour in case the file already exists at this
                output_prefix. Options are
                **`[w]rite`**: (default) overwrites file/delete the folder if it already exists;
                **`[a]ppend`**: appends to the file/folder if it exists, otherwise creates a new file;
                **`[x]`** or **`fail`**: fails if file/folder already exists;
            tar: if True creates a tar archive instead of a folder.

        """
        super().__init__()

        # Shorthands for mode
        if mode == "w":
            mode = "write"
        elif mode == "a":
            mode = "append"
        elif mode == "x":
            mode = "fail"

        if not ((mode == "write") or (mode == "append") or (mode == "fail")):
            raise ValueError(
                "Mode not recognized: should be one of `[w]rite`, `[a]ppend` or "
                "`[x]`(fail)."
            )

        if tar is True:
            file_exists = _path.exists(output_prefix + ".tar")
        else:
            if output_prefix[-1] != "/":
                output_prefix = output_prefix + "/"
            file_exists = _path.exists(output_prefix)

        if file_exists and mode == "fail":
            raise ValueError(
                "Output file/folder already exists. Either delete it manually or "
                "change `output_prefix`."
            )

        dir_name = _path.dirname(output_prefix)
        if dir_name != "":
            os.makedirs(dir_name, exist_ok=True)

        self._prefix = output_prefix
        self._file_mode: FileModeT = mode

        self._save_every = save_every
        self._old_step = 0
        self._steps_notsaved = 0
        self._init = False

        self._runtime_taken = 0.0

        # tar
        self._tar = tar
        self._tar_file = None
        self._closed = False

        self._file_step = 0

    def _init_output(self):
        if self._is_master_process:
            if self._tar:
                self._create_tar_file()
            else:
                self._check_output_folder()
        self._init = True

    def _create_tar_file(self):
        if self._tar_file is None:
            self._tar_file = tarfile.TarFile(self._prefix + ".tar", self._file_mode[0])  # type: ignore
            self._file_step = 0
            if self._file_mode == "append":
                files = self._tar_file.getnames()
                file_numbers = [int(file[:-6]) for file in files]
                file_numbers.sort()
                self._file_step = file_numbers[-1] + 1

    def _check_output_folder(self):
        self._file_step = 0
        if self._file_mode == "write":
            for file in glob.glob(self._prefix + "*.mpack"):
                os.remove(file)
            os.makedirs(self._prefix, exist_ok=True)
        elif self._file_mode == "append":
            files = glob.glob(self._prefix + "*.mpack")
            file_numbers = [int(_path.basename(file)[:-6]) for file in files]
            file_numbers.sort()
            self._file_step = file_numbers[-1] + 1

    def close(self):
        if not self._closed and self._tar_file is not None:
            self._tar_file.close()
            self._closed = True

    def __call__(
        self,
        step: int,
        item: dict[str, Any],
        variational_state: "VariationalState | None" = None,
    ):
        old_step = self._old_step

        if self._steps_notsaved % self._save_every == 0 or step == old_step - 1:
            self._save_variables(variational_state)

        self._old_step = step
        self._steps_notsaved += 1

    def _save_variables(self, variational_state):
        if self._init is False:
            self._init_output()

        _time = time.time()
        binary_data = serialization.to_bytes(
            extract_replicated(variational_state.variables)
        )

        if self._is_master_process:
            if self._tar:
                save_binary_to_tar(
                    self._tar_file, binary_data, str(self._file_step) + ".mpack"
                )
            else:
                with open(self._prefix + str(self._file_step) + ".mpack", "wb") as f:
                    f.write(binary_data)

        self._file_step += 1
        self._runtime_taken += time.time() - _time

    def __del__(self):
        if hasattr(self, "_closed"):
            self.close()

    def flush(self, variational_state=None):
        pass

    def __repr__(self):
        return f"TarLog('{self._prefix}', mode={self._file_mode})"

    def __str__(self):
        _str = self.__repr__()
        _str = _str + f"\n  Runtime cost: {self._runtime_taken}"
        return _str



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import flax
import jax
from jax import numpy as jnp

from netket import jax as nkjax
from netket import config
from netket.hilbert import DiscreteHilbert
from netket.sampler import Sampler, SamplerState
from netket.utils.types import PRNGKeyT, DType


class ARDirectSamplerState(SamplerState):
    key: PRNGKeyT
    """state of the random number generator."""

    def __init__(self, key):
        self.key = key
        super().__init__()


class ARDirectSampler(Sampler):
    r"""
    Direct sampler for autoregressive neural networks.

    This sampler only works with Flax models.
    This flax model must expose a specific method, `model.conditional`, which given
    a batch of samples and an index `i∈[0,self.hilbert.size]` must return the vector
    of partial probabilities at index `i` for the various (partial) samples provided.

    In short, if your model can be sampled according to a probability
    $ p(x) = p_1(x_1)p_2(x_2|x_1)\dots p_N(x_N|x_{N-1}\dots x_1) $ then
    `model.conditional(x, i)` should return $p_i(x)$.

    NetKet implements some autoregressive networks that can be used together with this
    sampler.
    """

    def __init__(
        self,
        hilbert: DiscreteHilbert,
        machine_pow: None = None,
        dtype: DType = None,
    ):
        """
        Construct an autoregressive direct sampler.

        Args:
            hilbert: The Hilbert space to sample.
            dtype: The dtype of the states sampled (default = np.float64).

        Note:
            `ARDirectSampler.machine_pow` has no effect. Please set the model's `machine_pow` instead.
        """

        if machine_pow is not None:
            raise ValueError(
                "ARDirectSampler.machine_pow should not be used. Modify the model `machine_pow` directly."
            )

        if hilbert.constrained:
            raise ValueError(
                "Only unconstrained Hilbert spaces can be sampled autoregressively with "
                "this sampler. To sample constrained spaces, you must write your own (do get in "
                "touch with us. We are interested!)"
            )

        super().__init__(hilbert, machine_pow=2, dtype=dtype)
        # ensure machine_pow is a float, as it can be sometimes used around...
        self.machine_pow = float(self.machine_pow)

    @property
    def is_exact(sampler):
        """
        Returns `True` because the sampler is exact.

        The sampler is exact if all the samples are exactly distributed according to the
        chosen power of the variational state, and there is no correlation among them.
        """
        return True

    def _init_cache(self, model, σ, key):
        variables = model.init(key, σ, 0, method=model.conditional)
        cache = variables.get("cache")
        return cache

    def _init_state(self, model, variables, key):
        return ARDirectSamplerState(key=key)

    def _reset(self, model, variables, state):
        return state

    @partial(
        jax.jit, static_argnames=("model", "chain_length", "return_log_probabilities")
    )
    def _sample_chain(
        self,
        model,
        variables,
        state,
        chain_length,
        return_log_probabilities: bool = False,
    ):
        if return_log_probabilities:
            raise NotImplementedError(
                "return_log_probability is not implemented for ARDirectSampler"
            )

        if "cache" in variables:
            variables, _ = flax.core.pop(variables, "cache")
        variables_no_cache = variables

        def scan_fun(carry, index):
            σ, cache, key = carry
            if cache:
                variables = {**variables_no_cache, "cache": cache}
            else:
                variables = variables_no_cache
            new_key, key = jax.random.split(key)

            p, mutables = model.apply(
                variables,
                σ,
                index,
                method=model.conditional,
                mutable=["cache"],
            )
            cache = mutables.get("cache")

            local_states = jnp.asarray(self.hilbert.local_states, dtype=self.dtype)
            new_σ = nkjax.batch_choice(key, local_states, p)
            σ = σ.at[:, index].set(new_σ)

            return (σ, cache, new_key), None

        new_key, key_init, key_scan = jax.random.split(state.key, 3)

        # Initialize a buffer for `σ` before generating a batch of samples
        # The result should not depend on its initial content
        σ = jnp.zeros(
            (self.n_batches * chain_length, self.hilbert.size),
            dtype=self.dtype,
        )

        if config.netket_experimental_sharding:
            σ = jax.lax.with_sharding_constraint(
                σ, jax.sharding.PositionalSharding(jax.devices()).reshape(-1, 1)
            )

        # Initialize `cache` before generating a batch of samples,
        # even if `variables` is not changed and `reset` is not called
        cache = self._init_cache(model, σ, key_init)
        if cache:
            variables = {**variables_no_cache, "cache": cache}
        else:
            variables = variables_no_cache

        indices = jnp.arange(self.hilbert.size)
        indices = model.apply(variables, indices, method=model.reorder)
        (σ, _, _), _ = jax.lax.scan(scan_fun, (σ, cache, key_scan), indices)
        σ = σ.reshape((self.n_batches, chain_length, self.hilbert.size))

        new_state = state.replace(key=new_key)
        return σ, new_state
